import re
from argparse import ArgumentParser, HelpFormatter
from contextlib import nullcontext
from functools import partial
from io import StringIO
from os.path import join
from runpy import run_module
from shlex import split
from sys import argv, executable, stderr
from tempfile import TemporaryDirectory
import pytest
import ubelt as ub
from line_profiler import LineProfiler
from line_profiler.cli_utils import add_argument


@pytest.fixture
def parser():
    """
    Argument parser with the following boolean flags:

    -f, -F, --foo -> foo
    -b, --bar -> bar
    -B, --baz -> baz
    --no-spam -> spam (negated)
    --ham -> ham
    -c -> c
    """
    parser = ArgumentParser(
        formatter_class=partial(HelpFormatter,
                                max_help_position=float('inf'),
                                width=float('inf')))
    # Normal boolean flag (w/2 short forms)
    # -> adds 3 actions (long, short, long-negated)
    add_argument(parser, '-f', '-F', '--foo', action='store_true')
    # Boolean flag w/o parenthetical remark in help text
    # -> adds 3 actions (long, short, long-negated)
    add_argument(parser, '-b', '--bar', action='store_true', help='Set `bar`')
    # Boolean flag w/parenthetical remark in help text
    # -> adds 3 actions (long, short, long-negated)
    add_argument(parser, '-B', '--baz',
                 action='store_true', help='Set `baz` (BAZ)')
    # Negative boolean flag
    # -> adds 1 action (long-negated)
    add_argument(parser, '--no-spam',
                 action='store_false', dest='spam', help='Set `spam` to false')
    # Boolean flag w/o short form
    # -> adds 2 actions (long, long-negated)
    add_argument(parser, '--ham', action='store_true', help='Set `ham`')
    # Short-form-only boolean flag
    # -> adds 1 action (short)
    add_argument(parser, '-e',
                 action='store_true', dest='eggs', help='Set `eggs`')
    yield parser


def test_boolean_argument_help_text(parser):
    """
    Test the help texts generated from boolean arguments added by
    `line_profiler.cli_utils.add_argument(action=...)`.
    """
    assert len(parser._actions) == 14  # One extra option from `--help`
    with StringIO() as sio:
        parser.print_help(sio)
        help_text = sio.getvalue()
    matches = partial(re.search, string=help_text, flags=re.MULTILINE)
    assert matches(r'^  --foo \[.*\] +'
                   + re.escape('(Short forms: -f, -F)')
                   + '$')
    assert matches(r'^  --bar \[.*\] +'
                   + re.escape('Set `bar` (Short form: -b)')
                   + '$')
    assert matches(r'^  --baz \[.*\] +'
                   + re.escape('Set `baz` (BAZ; short form: -B)')
                   + '$')
    assert matches(r'^  --no-spam \[.*\] +'
                   + re.escape('Set `spam` to false')
                   + '$')
    assert matches(r'^  --ham \[.*\] +'
                   + re.escape('Set `ham`')
                   + '$')
    assert matches(r'^  -e +'
                   + re.escape('Set `eggs`')
                   + '$')


@pytest.mark.parametrize(
    ('args', 'foo', 'bar', 'baz', 'spam', 'ham', 'eggs', 'expect_error'),
    [('--foo q', *((None,) * 6), True),  # Can't parse `q` into boolean
     ('-fbB'  # Test short-flag concatenation
      ' --ham=',  # Empty string -> set to false
      True, True, True, None, False, None, False),
     ('--foo'  # No-arg -> set to true
      ' --bar=0'  # Falsy arg -> set to false
      ' --no-baz'  # No-arg (negated flag) -> set to false
      ' --no-spam=no'  # Falsy arg (negated flag) -> set to true
      ' --ham=on'  # Truey arg -> set to true
      ' -e',  # No-arg -> set to true
      True, False, False, True, True, True, False)])
def test_boolean_argument_parsing(
        parser, capsys, args, foo, bar, baz, spam, ham, eggs, expect_error):
    """
    Test the handling of boolean flags.
    """
    if expect_error:
        ctx = pytest.raises(SystemExit)
        match_stderr = 'usage: .* error: argument'
    else:
        ctx = nullcontext()
        match_stderr = '^$'
    with ctx:
        result = vars(parser.parse_args(split(args)))
    stderr = capsys.readouterr().err
    assert re.match(match_stderr, stderr, flags=re.DOTALL)
    if expect_error:
        return
    expected = dict(foo=foo, bar=bar, baz=baz, spam=spam, ham=ham, eggs=eggs)
    assert result == expected


def test_cli():
    """
    Test command line interaction with kernprof and line_profiler.

    References:
        https://github.com/pyutils/line_profiler/issues/9

    CommandLine:
        xdoctest -m ./tests/test_cli.py test_cli
    """
    # Create a dummy source file
    code = ub.codeblock(
        '''
        @profile
        def my_inefficient_function():
            a = 0
            for i in range(10):
                a += i
                for j in range(10):
                    a += j

        if __name__ == '__main__':
            my_inefficient_function()
        ''')
    with TemporaryDirectory() as tmp_dpath:
        tmp_src_fpath = join(tmp_dpath, 'foo.py')
        with open(tmp_src_fpath, 'w') as file:
            file.write(code)

        # Run kernprof on it
        info = ub.cmd(f'kernprof -l {tmp_src_fpath}', verbose=3, cwd=tmp_dpath)
        assert info['ret'] == 0

        tmp_lprof_fpath = join(tmp_dpath, 'foo.py.lprof')
        tmp_lprof_fpath

        info = ub.cmd(f'{executable} -m line_profiler {tmp_lprof_fpath}',
                      cwd=tmp_dpath, verbose=3)
        assert info['ret'] == 0
        # Check for some patterns that should be in the output
        assert '% Time' in info['out']
        assert '7       100' in info['out']


def test_multiple_lprof_files(capsys):
    """
    Test that we can aggregate profiling results with
    ``python -m line_profiler``.
    """
    def sum_n(n: int) -> int:
        x = 0
        for n in range(1, n + 1):
            x += n  # Loop: sum_n
        return x  # Return: sum_n

    def sum_nsq(n: int) -> int:
        x = 0
        for n in range(1, n + 1):
            x += n * n  # Loop: sum_nsq
        return x  # Return: sum_nsq

    profs = {0: LineProfiler(sum_n),
             1: LineProfiler(sum_nsq),
             2: LineProfiler(sum_n, sum_nsq)}

    with TemporaryDirectory() as tmp_dpath:
        # Write several profiling output files
        stats_files = []
        nhits = {}
        for i, (func, n, expected) in enumerate([
                (sum_n, 10, 10 * 11 // 2),
                (sum_nsq, 20, 20 * 21 * 41 // 6),
                (sum_n, 30, 30 * 31 // 2)]):
            prof = profs[i]
            with prof:
                assert func(n) == expected
            stats = join(tmp_dpath, f'{i}.lprof')
            stats_files.append(stats)
            prev_loop, prev_return = nhits.get(func, (0, 0))
            nhits[func] = prev_loop + n, prev_return + 1
            prof.dump_stats(stats)

        old_argv = argv.copy()
        argv[:] = ['line_profiler', *stats_files]
        try:
            run_module('line_profiler', run_name='__main__', alter_sys=True)
        finally:
            argv[:] = old_argv

    # View them and check the output
    checks = {}
    out, err = capsys.readouterr()
    print(out, end='')
    print(err, end='', file=stderr)
    for func in sum_n, sum_nsq:
        for comment, n in zip(['Loop', 'Return'], nhits[func]):
            checks[f'  # {comment}: {func.__name__}'] = n
    for line in out.splitlines():
        try:
            suffix, = (suffix for suffix in checks if line.endswith(suffix))
        except ValueError:  # No match
            continue
        assert int(line.split()[1]) == checks.pop(suffix)


def test_version_agreement():
    """
    Ensure that line_profiler and kernprof have the same version info
    """
    info1 = ub.cmd(f'{executable} -m line_profiler --version')
    info2 = ub.cmd(f'{executable} -m kernprof --version')

    if info1['ret'] != 0:
        print(f'Error querying line-profiler version: {info1}')

    if info2['ret'] != 0:
        print(f'Error querying kernprof version: {info2}')

    # Strip local version suffixes
    version1 = info1['out'].strip().split('+')[0]
    version2 = info2['out'].strip().split('+')[0]

    if version2 != version1:
        raise AssertionError(
            'Version Mismatch: kernprof and line_profiler must be in sync. '
            f'kernprof.line_profiler = {version1}. '
            f'kernprof.__version__ = {version2}. '
        )


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/line_profiler/tests/test_cli.py
    """
    test_version_agreement()



"""
Tools for eagerly pre-importing everything as specified in
``line_profiler.autoprof.run(prof_mod=...)``.
"""
import ast
import functools
import itertools
from collections import namedtuple
from collections.abc import Collection
from keyword import iskeyword
from importlib.util import find_spec
from pkgutil import walk_packages
from textwrap import dedent, indent as indent_
from warnings import warn
from .util_static import (
    modname_to_modpath, modpath_to_modname, package_modpaths)


__all__ = ('is_dotted_path', 'split_dotted_path',
           'resolve_profiling_targets', 'write_eager_import_module')


def is_dotted_path(obj):
    """
    Example:
        >>> assert not is_dotted_path(object())
        >>> assert is_dotted_path('foo')
        >>> assert is_dotted_path('foo.bar')
        >>> assert not is_dotted_path('not an identifier')
        >>> assert not is_dotted_path('keyword.return.not.allowed')
    """
    if not (isinstance(obj, str) and obj):
        return False
    for chunk in obj.split('.'):
        if iskeyword(chunk) or not chunk.isidentifier():
            return False
    return True


def get_expression(obj):
    """
    Example:
        >>> assert not get_expression(object())
        >>> assert not get_expression('')
        >>> assert not get_expression('foo; bar')
        >>> assert get_expression('foo')
        >>> assert get_expression('lambda x: x')
        >>> assert not get_expression('def foo(x): return x')
    """
    if not (isinstance(obj, str) and obj):
        return None
    try:
        return ast.parse(obj, mode='eval')
    except SyntaxError:
        return None


def split_dotted_path(dotted_path, static=True):
    """
    Arguments:
        dotted_path (str):
            Dotted path indicating an import target (module, package, or
            a ``from ... import ...``-able name under that), or an
            object accessible via (chained) attribute access thereon
        static (bool):
            Whether to use static analysis (true) or the import system
            (false) to resolve targets

    Returns:
        module, target (tuple[str, Union[str, None]]):

        * ``module``: dotted path indicating the module that should be
          imported
        * ``target``: dotted path indicating the chained-attribute
          access target on the imported module corresponding to
          ``dotted_path``;
          if the import is just a module, this is set to
          :py:const:`None`

    Raises:
        TypeError
            If ``dotted_path`` is not a dotted path (Python identifiers
            joined by periods)
        ModuleNotFoundError
            If a matching module cannot be found

    Note:
        ``static=False`` can cause the ancestor module objects of
        ``dotted_path`` to be imported (and hence alter the state of
        :py:data:`sys.modules`.

    Example:
        >>> split_dotted_path('importlib.util.find_spec')
        ('importlib.util', 'find_spec')
        >>> split_dotted_path('importlib.util')
        ('importlib.util', None)
        >>> split_dotted_path('importlib.abc.Loader.exec_module')
        ('importlib.abc', 'Loader.exec_module')
        >>> split_dotted_path(  # doctest: +NORMALIZE_WHITESPACE
        ...     'not a dotted path')
        Traceback (most recent call last):
          ...
        TypeError: dotted_path = 'not a dotted path':
        expected a dotted path (string of period-joined identifiers)
        >>> split_dotted_path(  # doctest: +NORMALIZE_WHITESPACE
        ...     'foo.bar.baz')
        Traceback (most recent call last):
          ...
        ModuleNotFoundError: dotted_path = 'foo.bar.baz':
        none of the below looks like an importable module:
        ['foo.bar.baz', 'foo.bar', 'foo']
    """
    if not is_dotted_path(dotted_path):
        raise TypeError(f'dotted_path = {dotted_path!r}: '
                        'expected a dotted path '
                        '(string of period-joined identifiers)')
    chunks = dotted_path.split('.')
    checked_locs = []
    check = modname_to_modpath if static else find_spec
    for slicing_point in range(len(chunks), 0, -1):
        module = '.'.join(chunks[:slicing_point])
        target = '.'.join(chunks[slicing_point:]) or None
        try:
            result = check(module)
        except ImportError:
            result = None
        if result is None:
            checked_locs.append(module)
            continue
        return module, target
    raise ModuleNotFoundError(f'dotted_path = {dotted_path!r}: '
                              'none of the below looks like an importable '
                              f'module: {checked_locs!r}')


def strip(s):
    return dedent(s).strip('\n')


class LoadedNameFinder(ast.NodeVisitor):
    """
    Find the names loaded in an AST.  A name is considered to be loaded
    if it appears with the context :py:class:`ast.Load()` and is not an
    argument of any surrounding function-definition contexts
    (``def func(...): ...``, ``async def func(...): ...``, or
    ``lambda ...: ...``).

    Example:
        >>> import ast
        >>>
        >>>
        >>> module = '''
        ... def foo(x, **k):
        ...     def bar(y, **z):
        ...         pass
        ...
        ...     return bar(x, **{**k, 'baz': foobar})
        ...
        ... spam = lambda x, *y, **z: (x, y, z, a)
        ...
        ... str('ham')
        ... '''
        >>> names = LoadedNameFinder.find(ast.parse(module))
        >>> assert names == {'bar', 'foobar', 'a', 'str'}, names
    """
    def __init__(self):
        self.names = set()
        self.contexts = []

    def visit_Name(self, node):
        if not isinstance(node.ctx, ast.Load):
            return
        name = node.id
        if not any(name in ctx for ctx in self.contexts):
            self.names.add(node.id)

    def _visit_func_def(self, node):
        args = node.args
        arg_names = {
            arg.arg
            for arg_list in (args.posonlyargs, args.args, args.kwonlyargs)
            for arg in arg_list}
        if args.vararg:
            arg_names.add(args.vararg.arg)
        if args.kwarg:
            arg_names.add(args.kwarg.arg)
        self.contexts.append(arg_names)
        self.generic_visit(node)
        self.contexts.pop()

    visit_FunctionDef = visit_AsyncFunctionDef = visit_Lambda = _visit_func_def

    @classmethod
    def find(cls, node):
        finder = cls()
        finder.visit(node)
        return finder.names


def propose_names(prefixes):
    """
    Generate names based on prefixes.

    Arguments:
        prefixes (Collection[str]):
            String identifier prefixes

    Yields:
        name (str):
            String identifier

    Example:
        >>> import itertools
        >>>
        >>>
        >>> list(itertools.islice(propose_names(['func', 'f', 'foo']),
        ...                       10))  # doctest: +NORMALIZE_WHITESPACE
        ['func', 'f', 'foo',
         'func_0', 'f0', 'foo_0',
         'func_1', 'f1', 'foo_1',
         'func_2']
    """
    prefixes = list(dict.fromkeys(prefixes))  # Preserve order
    if not all(is_dotted_path(p) and '.' not in p for p in prefixes):
        raise TypeError(f'prefixes = {prefixes!r}: '
                        'expected string identifiers')
    # Yield all the provided prefixes
    yield from prefixes
    # Yield the prefixes in order with numeric suffixes
    prefixes_and_patterns = [
        (prefix, ('{}{}' if len(prefix) == 1 else '{}_{}').format)
        for prefix in prefixes]
    for i in itertools.count():
        for prefix, pattern in prefixes_and_patterns:
            yield pattern(prefix, i)


def resolve_profiling_targets(dotted_paths, static=True, recurse=False):
    """
    Arguments:
        dotted_paths (Collection[str]):
            Dotted paths (strings of period-joined identifiers)
            indicating what should be profiled
        static (bool):
            Whether to use static analysis (true) or the import system
            (false) to resolve targets
        recurse (Union[Collection[str], bool]):
            Dotted paths (strings of period-joined identifiers)
            indicating the profiling targets that should be recursed
            into if they are packages;
            can also be a boolean value, indicating:

            :py:const:`True`
                Recurse into any entry in ``dotted_paths`` that is a
                package
            :py:const:`False`
                Don't recurse into any entry

    Returns:
        result (ResolvedResult):
            3-named-tuple with the following fields/items:

            ``.targets`` (dict[str, set[str | None]]):
                Mapping from module names to the names of the attributes
                therein that should be passed to the profiler;
                if the attribute name is :py:const:`None`, the whole
                module is to be passed to the profiler
            ``.indirect`` (set[str]):
                Set of subpackage/-module names included only via
                ``recurse``-ing into packages (i.e. not directly in
                either ``dotted_paths`` or ``recurse``)
            ``.unresolved`` (list[str]):
                List of unresolved profiling targets, i.e. those which
                cannot be resolved into a module part and an attribute
                part by :py:func:`~.split_dotted_path`

    Note:
        ``static=False`` can cause the ancestor module objects of
        ``dotted_paths`` and ``recurse`` to be imported (and hence alter
        the state of :py:data:`sys.modules`.
    """
    def walk_packages_static(pkg):
        # Note: this probably can't handle namespace packages
        path = modname_to_modpath(pkg)
        if not path:
            return
        for subpath in package_modpaths(path, with_pkg=True):
            submod = modpath_to_modname(subpath)
            if submod and is_dotted_path(submod):
                yield submod

    def walk_packages_import_sys(pkg):
        spec = find_spec(pkg)
        if not spec:
            return
        paths = spec.submodule_search_locations or []
        if not paths:
            return
        for info in walk_packages(paths, prefix=pkg + '.'):
            if is_dotted_path(info.name):
                yield info.name

    dotted_paths = set(dotted_paths)
    if isinstance(recurse, Collection):
        recurse = set(recurse)
    else:
        recurse = dotted_paths if recurse else set()
    dotted_paths |= recurse
    indirect_submods = set()

    all_targets = {}
    unknown_locs = []
    split_path = functools.partial(split_dotted_path, static=static)
    walk = walk_packages_static if static else walk_packages_import_sys
    for path in sorted(set(dotted_paths)):
        try:
            module, target = split_path(path)
        except ModuleNotFoundError:
            unknown_locs.append(path)
            continue
        all_targets.setdefault(module, set()).add(target)
        if path in recurse and target is None:
            for submod in walk(path):
                all_targets.setdefault(submod, set()).add(None)
                indirect_submods.add(submod)
    indirect_submods -= dotted_paths
    return ResolvedResult(all_targets, indirect_submods, unknown_locs)


def write_eager_import_module(dotted_paths, stream=None, *,
                              static=True,
                              recurse=False,
                              adder='profile.add_imported_function_or_module',
                              indent='    '):
    r"""
    Write a module which autoprofiles all its imports.

    Arguments:
        dotted_paths (Collection[str]):
            Dotted paths (strings of period-joined identifiers)
            indicating what should be profiled
        stream (Union[TextIO, None]):
            Optional text-mode writable file object to which to write
            the module
        static (bool):
            Whether to use static analysis (true) or the import system
            (false) to resolve targets
        recurse (Union[Collection[str], bool]):
            Dotted paths (strings of period-joined identifiers)
            indicating the profiling targets that should be recursed
            into if they are packages;
            can also be a boolean value, indicating:

            :py:const:`True`
                Recurse into any entry in ``dotted_paths`` that is a
                package
            :py:const:`False`
                Don't recurse into any entry
        adder (str):
            Single-line string ``ast.parse(mode='eval')``-able to a
            single expression, indicating the callable (which is assumed
            to exist in the builtin namespace by the time the module is
            executed) to be called to add the profiling target
        indent (str):
            Single-line, non-empty whitespace string to indent the
            output with

    Side effects:
        * ``stream`` (or :py:data:`sys.stdout` if :py:const:`None`)
          written to
        * Warning issued if the module can't be located for one or more
          dotted paths

    Raises:
        TypeError
            If ``adder`` and ``indent`` are not strings, **OR**
            if ``dotted_paths`` is not a collection of dotted paths
        ValueError
            If ``adder`` is a non-single-line string or is not parsable
            to a single expression, **OR**
            if ``indent`` isn't single-line, non-empty, and whitespace

    Example:
        >>> import io
        >>> import textwrap
        >>> import warnings
        >>>
        >>>
        >>> def strip(s):
        ...     return textwrap.dedent(s).strip('\n')
        ...
        >>>
        >>> with warnings.catch_warnings(record=True) as record:
        ...     with io.StringIO() as sio:
        ...         write_eager_import_module(
        ...             ['importlib.util',
        ...              'foo.bar',
        ...              'importlib.abc.Loader.exec_module',
        ...              'importlib.abc.Loader.find_module'],
        ...             sio)
        ...         written = strip(sio.getvalue())
        ...
        >>> assert written == strip('''
        ... add = profile.add_imported_function_or_module  # noqa: F821
        ... failures = []
        ...
        ...
        ... try:
        ...     import importlib.abc as module
        ... except ImportError:
        ...     pass
        ... else:
        ...     try:
        ...         add(module.Loader.exec_module)
        ...     except AttributeError:
        ...         failures.append('importlib.abc.Loader.exec_module')
        ...     try:
        ...         add(module.Loader.find_module)
        ...     except AttributeError:
        ...         failures.append('importlib.abc.Loader.find_module')
        ...
        ... try:
        ...     import importlib.util as module
        ... except ImportError:
        ...     failures.append('importlib.util')
        ... else:
        ...     add(module)
        ...
        ...
        ... if failures:
        ...     import warnings
        ...
        ...     msg = '{} target{} cannot be imported: {!r}'.format(
        ...         len(failures),
        ...         '' if len(failures) == 1 else 's',
        ...         failures)
        ...     warnings.warn(msg, stacklevel=2)
        ... '''), written
        >>> assert len(record) == 1
        >>> assert (record[0].message.args[0]
        ...         == ("1 import target cannot be resolved: "
        ...             "['foo.bar']"))

    Note:
        ``static=False`` can cause the ancestor module objects of
        ``dotted_paths`` and ``recurse`` to be imported (and hence alter
        the state of :py:data:`sys.modules`.
    """
    if not isinstance(adder, str):
        AdderError = TypeError
    elif len(adder.splitlines()) != 1:
        AdderError = ValueError
    else:
        expr = get_expression(adder)
        if expr:
            AdderError = None
        else:
            AdderError = ValueError
    if AdderError:
        raise AdderError(f'adder = {adder!r}: '
                         'expected a single-line string parsable to a single '
                         'expression')
    if not isinstance(indent, str):
        IndentError = TypeError
    elif len(indent.splitlines()) == 1 and indent.isspace():
        IndentError = None
    else:
        IndentError = ValueError
    if IndentError:
        raise IndentError(f'indent = {indent!r}: '
                          'expected a single-line non-empty whitespace string')

    # Get the names loaded by `adder`;
    # these names are not allowed in the namespace
    forbidden_names = LoadedNameFinder.find(expr)
    # We need three free names:
    # - One for `adder`
    # - One for a list of failed targets
    # - One for the imported module
    adder_name = next(
        name for name in propose_names(['add', 'add_func', 'a', 'f'])
        if name not in forbidden_names)
    forbidden_names.add(adder_name)
    failures_name = next(
        name
        for name in propose_names(['failures', 'failed_targets', 'f', '_'])
        if name not in forbidden_names)
    forbidden_names.add(failures_name)
    module_name = next(
        name for name in propose_names(['module', 'mod', 'imported', 'm', '_'])
        if name not in forbidden_names)

    # Figure out the import targets to profile
    resolved = resolve_profiling_targets(
        dotted_paths, static=static, recurse=recurse)

    # Warn against failed imports
    if resolved.unresolved:
        msg = '{} import target{} cannot be resolved: {!r}'.format(
            len(resolved.unresolved),
            '' if len(resolved.unresolved) == 1 else 's',
            resolved.unresolved)
        warn(msg, stacklevel=2)

    # Do the imports and add them with `adder`
    write = functools.partial(print, file=stream)
    write(f'{adder_name} = {adder}  # noqa: F821\n{failures_name} = []')
    for i, (module, targets) in enumerate(resolved.targets.items()):
        assert targets
        # Write one more empty line so that the imports are separated
        # from the preambles by 2 lines
        if not i:
            write()
        # Allow arbitrary errors for modules that are only added
        # indirectly (via descent/recursion)
        if module in resolved.indirect:
            allowed_error = 'Exception'
        else:
            allowed_error = 'ImportError'
        # Is the module itself a direct target?
        try:
            targets.remove(None)
        except KeyError:  # Not found
            profile_whole_module = False
        else:
            profile_whole_module = True
        if profile_whole_module:
            on_error = f'{failures_name}.append({module!r})'
        else:
            on_error = 'pass'
        write('\n'
              + strip(f"""
            try:
            {indent}import {module} as {module_name}
            except {allowed_error}:
            {indent}{on_error}
            else:
                """))
        chunks = []
        if profile_whole_module:
            chunks.append(f'{adder_name}({module_name})')
        for target in sorted(targets):
            path = f'{module}.{target}'
            chunks.append(strip(f"""
            try:
            {indent}{adder_name}({module_name}.{target})
            except AttributeError:
            {indent}{failures_name}.append({path!r})
            """))
        for chunk in chunks:
            write(indent_(chunk, indent))
    # Issue a warning if any of the targets doesn't exist
    if resolved.targets:
        write('\n')
        write(strip(f"""
        if {failures_name}:
        {indent}import warnings

        {indent}msg = '{{}} target{{}} cannot be imported: {{!r}}'.format(
        {indent * 2}len({failures_name}),
        {indent * 2}'' if len({failures_name}) == 1 else 's',
        {indent * 2}{failures_name})
        {indent}warnings.warn(msg, stacklevel=2)
        """))


ResolvedResult = namedtuple('ResolvedResult',
                            ('targets', 'indirect', 'unresolved'))



from functools import partial, partialmethod, cached_property
from types import FunctionType, MethodType, ModuleType
from typing import overload, Any, Literal, TypeVar, TYPE_CHECKING

if TYPE_CHECKING:  # Stub-only annotations
    from ..profiler_mixin import CLevelCallable, CythonCallable
    from ..scoping_policy import ScopingPolicy, ScopingPolicyDict




@overload
def add_imported_function_or_module(
        self, item: CLevelCallable | Any,
        scoping_policy: ScopingPolicy | str | ScopingPolicyDict | None = None,
        wrap: bool = False) -> Literal[0]:
    ...


@overload
def add_imported_function_or_module(
        self,
        item: (FunctionType | CythonCallable
               | type | partial | property | cached_property
               | MethodType | staticmethod | classmethod | partialmethod
               | ModuleType),
        scoping_policy: ScopingPolicy | str | ScopingPolicyDict | None = None,
        wrap: bool = False) -> Literal[0, 1]:
    ...



import io
from functools import cached_property, partial, partialmethod
from os import PathLike
from types import FunctionType, ModuleType
from typing import (TYPE_CHECKING,
                    overload,
                    Callable, Mapping,
                    Literal, Self,
                    Protocol, TypeVar)
try:
    from typing import (  # type: ignore[attr-defined]  # noqa: F401
        ParamSpec)
except ImportError:
    from typing_extensions import ParamSpec  # noqa: F401
from _typeshed import Incomplete
from ._line_profiler import (LineProfiler as CLineProfiler,
                             LineStats as CLineStats)
from .profiler_mixin import ByCountProfilerMixin, CLevelCallable
from .scoping_policy import ScopingPolicy, ScopingPolicyDict

if TYPE_CHECKING:
    from .profiler_mixin import UnparametrizedCallableLike


T = TypeVar('T')
T_co = TypeVar('T_co', covariant=True)
PS = ParamSpec('PS')


def get_column_widths(
    config: bool | str | PathLike[str] | None = False) -> Mapping[
        Literal['line', 'hits', 'time', 'perhit', 'percent'], int]:
    ...


def load_ipython_extension(ip) -> None:
    ...


class _StatsLike(Protocol):
    timings: Mapping[tuple[str, int, str],  # funcname, lineno, filename
                     list[tuple[int, int, int]]]  # lineno, nhits, time
    unit: float


class LineStats(CLineStats):
    def to_file(self, filename: PathLike[str] | str) -> None:
        ...

    def print(self, stream: Incomplete | None = None, **kwargs) -> None:
        ...

    @classmethod
    def from_files(cls, file: PathLike[str] | str, /,
                   *files: PathLike[str] | str) -> Self:
        ...

    @classmethod
    def from_stats_objects(cls, stats: _StatsLike, /,
                           *more_stats: _StatsLike) -> Self:
        ...

    def __repr__(self) -> str:
        ...

    def __eq__(self, other) -> bool:
        ...

    def __add__(self, other: _StatsLike) -> Self:
        ...

    def __iadd__(self, other: _StatsLike) -> Self:
        ...


class LineProfiler(CLineProfiler, ByCountProfilerMixin):
    @overload
    def __call__(self,  # type: ignore[overload-overlap]
                 func: CLevelCallable) -> CLevelCallable:
        ...

    @overload
    def __call__(  # type: ignore[overload-overlap]
        self, func: UnparametrizedCallableLike,
    ) -> UnparametrizedCallableLike:
        ...

    @overload
    def __call__(self,  # type: ignore[overload-overlap]
                 func: type[T]) -> type[T]:
        ...

    @overload
    def __call__(self,  # type: ignore[overload-overlap]
                 func: partial[T]) -> partial[T]:
        ...

    @overload
    def __call__(self, func: partialmethod[T]) -> partialmethod[T]:
        ...

    @overload
    def __call__(self, func: cached_property[T_co]) -> cached_property[T_co]:
        ...

    @overload
    def __call__(self,  # type: ignore[overload-overlap]
                 func: staticmethod[PS, T_co]) -> staticmethod[PS, T_co]:
        ...

    @overload
    def __call__(
        self, func: classmethod[type[T], PS, T_co],
    ) -> classmethod[type[T], PS, T_co]:
        ...

    # Fallback: just wrap the `.__call__()` of a generic callable

    @overload
    def __call__(self, func: Callable) -> Callable:
        ...

    def add_callable(
            self, func,
            guard: Callable[[FunctionType], bool] | None = None,
            name: str | None = None) -> Literal[0, 1]:
        ...

    def get_stats(self) -> LineStats:
        ...

    def dump_stats(self, filename) -> None:
        ...

    def print_stats(self,
                    stream: Incomplete | None = ...,
                    output_unit: Incomplete | None = ...,
                    stripzeros: bool = ...,
                    details: bool = ...,
                    summarize: bool = ...,
                    sort: bool = ...,
                    rich: bool = ...,
                    *,
                    config: str | PathLike[str] | bool | None = None) -> None:
        ...

    def add_module(
            self, mod: ModuleType, *,
            scoping_policy: (
                ScopingPolicy | str | ScopingPolicyDict | None) = None,
            wrap: bool = False) -> int:
        ...

    def add_class(
            self, cls: type, *,
            scoping_policy: (
                ScopingPolicy | str | ScopingPolicyDict | None) = None,
            wrap: bool = False) -> int:
        ...


def is_generated_code(filename):
    ...


def show_func(filename: str,
              start_lineno: int,
              func_name: str,
              timings: list[tuple[int, int, float]],
              unit: float,
              output_unit: float | None = None,
              stream: io.TextIOBase | None = None,
              stripzeros: bool = False,
              rich: bool = False,
              *,
              config: str | PathLike[str] | bool | None = None) -> None:
    ...


def show_text(stats,
              unit,
              output_unit: Incomplete | None = ...,
              stream: Incomplete | None = ...,
              stripzeros: bool = ...,
              details: bool = ...,
              summarize: bool = ...,
              sort: bool = ...,
              rich: bool = ...,
              *,
              config: str | PathLike[str] | bool | None = None) -> None:
    ...


load_stats = LineStats.from_files


def main():
    ...



"""
Global state initialized at import time.
Used for hidden arguments and developer features.
"""
import os
import sys
from types import ModuleType
from line_profiler import _logger


def _boolean_environ(
        envvar,
        truey=frozenset({'1', 'on', 'true', 'yes'}),
        falsy=frozenset({'0', 'off', 'false', 'no'}),
        default=False):
    r"""
    Args:
        envvar (str)
            Name for the environment variable to read from.
        truey (Collection[str])
            Values to be considered truey.
        falsy (Collection[str])
            Values to be considered falsy.
        default (bool)
            Default boolean value to resolve to.

    Returns:
        :py:data:`True`
            If the (case-normalized) environment variable is equal to
            any of ``truey``.
        :py:data:`False`
            If the (case-normalized) environment variable is equal to
            any of ``falsy``.
        ``default``
            Otherwise.

    Example:
        >>> from os import environ
        >>> from subprocess import run
        >>> from sys import executable
        >>> from textwrap import dedent
        >>>
        >>>
        >>> def resolve_in_subproc(value, default,
        ...                        envvar='MY_ENVVAR',
        ...                        truey=('foo',), falsy=('bar',)):
        ...     code = dedent('''
        ...     from {0.__module__} import {0.__name__}
        ...     print({0.__name__}({1!r}, {2!r}, {3!r}, {4!r}))
        ...     ''').strip('\n').format(_boolean_environ, envvar,
        ...                             tuple(truey), tuple(falsy),
        ...                             bool(default))
        ...     env = environ.copy()
        ...     env[envvar] = value
        ...     proc = run([executable, '-c', code],
        ...                capture_output=True, env=env, text=True)
        ...     proc.check_returncode()
        ...     return {'True': True,
        ...             'False': False}[proc.stdout.strip()]
        ...
        >>>
        >>> # Truey value
        >>> assert resolve_in_subproc('FOO', True) == True
        >>> assert resolve_in_subproc('FOO', False) == True
        >>> # Falsy value
        >>> assert resolve_in_subproc('BaR', True) == False
        >>> assert resolve_in_subproc('BaR', False) == False
        >>> # Mismatch -> fall back to default
        >>> assert resolve_in_subproc('baz', True) == True
        >>> assert resolve_in_subproc('baz', False) == False
    """
    # (TODO: migrate to `line_profiler.cli_utils.boolean()` after
    # merging #335)
    try:
        value = os.environ.get(envvar).casefold()
    except AttributeError:  # None
        return default
    non_default_values = falsy if default else truey
    if value in {v.casefold() for v in non_default_values}:
        return not default
    return default


# `kernprof` switches
DEBUG = _boolean_environ('LINE_PROFILER_DEBUG')
NO_EXEC = _boolean_environ('LINE_PROFILER_NO_EXEC')
KEEP_TEMPDIRS = _boolean_environ('LINE_PROFILER_KEEP_TEMPDIRS')
STATIC_ANALYSIS = _boolean_environ('LINE_PROFILER_STATIC_ANALYSIS')

# `line_profiler._line_profiler` switches
WRAP_TRACE = _boolean_environ('LINE_PROFILER_WRAP_TRACE')
SET_FRAME_LOCAL_TRACE = _boolean_environ('LINE_PROFILER_SET_FRAME_LOCAL_TRACE')
_MUST_USE_LEGACY_TRACE = not isinstance(
    getattr(sys, 'monitoring', None), ModuleType)
USE_LEGACY_TRACE = (
    _MUST_USE_LEGACY_TRACE
    or _boolean_environ('LINE_PROFILER_CORE',
                        # Also provide `coverage-style` aliases
                        truey={'old', 'legacy', 'ctrace'},
                        falsy={'new', 'sys.monitoring', 'sysmon'},
                        default=_MUST_USE_LEGACY_TRACE))

log = _logger.Logger('line_profiler', backend='auto')



from enum import auto
from types import FunctionType, ModuleType
from typing import overload, Literal, Callable, TypedDict
from .line_profiler_utils import StringEnum


class ScopingPolicy(StringEnum):
    EXACT = auto()
    CHILDREN = auto()
    DESCENDANTS = auto()
    SIBLINGS = auto()
    NONE = auto()

    @overload
    def get_filter(
            self,
            namespace: type | ModuleType,
            obj_type: Literal['func']) -> Callable[[FunctionType], bool]:
        ...

    @overload
    def get_filter(
            self,
            namespace: type | ModuleType,
            obj_type: Literal['class']) -> Callable[[type], bool]:
        ...

    @overload
    def get_filter(
            self,
            namespace: type | ModuleType,
            obj_type: Literal['module']) -> Callable[[ModuleType], bool]:
        ...

    @classmethod
    def to_policies(
        cls,
        policies: (str | 'ScopingPolicy' | 'ScopingPolicyDict'
                   | None) = None) -> '_ScopingPolicyDict':
        ...


ScopingPolicyDict = TypedDict('ScopingPolicyDict',
                              {'func': str | ScopingPolicy,
                               'class': str | ScopingPolicy,
                               'module': str | ScopingPolicy})
_ScopingPolicyDict = TypedDict('_ScopingPolicyDict',
                               {'func': str | ScopingPolicy,
                                'class': str | ScopingPolicy,
                                'module': str | ScopingPolicy})



"""
Test the interoperability between `LineProfiler` and other `sys` tracing
facilities (e.g. Python functions registered via `sys.settrace()`.

Notes
-----
- By the very nature of the tests in this test module, they override
  `sys` trace functions, and are thus largely opaque towards
  `coverage.py`.
- However, there effects are isolated since each test is run in a
  separate Python subprocess.
"""
import concurrent.futures
import functools
import inspect
import linecache
import os
import subprocess
import shlex
import sys
import time
import tempfile
import textwrap
import threading
import pytest
from ast import literal_eval
from contextlib import nullcontext
from io import StringIO
from types import FrameType, ModuleType
from typing import Any, Optional, Union, Callable, List, Literal
from line_profiler import LineProfiler


# Common utilities

DEBUG = False
USE_SYS_MONITORING = isinstance(getattr(sys, 'monitoring', None), ModuleType)

Event = Literal['call', 'line', 'return', 'exception', 'opcode']
TracingFunc = Callable[[FrameType, Event, Any], Union['TracingFunc', None]]


def strip(s: str) -> str:
    return textwrap.dedent(s).strip('\n')


def isolate_test_in_subproc(
        func: Optional[Callable] = None, debug: bool = DEBUG) -> Callable:
    """
    Run the test function with the supplied arguments in a subprocess so
    that it doesn't pollute the state of the current interpretor.
    If `debug` is true, run with `pytest` for more detailed traceback.

    Notes
    -----
    - Code is written to a tempfile and run in a subprocess.
    - The test function should be import-able from the top-level
      namespace of this file.
    - All the arguments should be `ast.literal_eval()`-able.
    - Beware of using fixtures for these tests.
    """
    if func is None:
        return functools.partial(isolate_test_in_subproc, debug=debug)

    def message(msg: str, header: str, *,
                short: bool = False, **kwargs) -> None:
        header = strip(header)
        if not header.endswith(':'):
            header += ':'
        kwargs['sep'] = '\n'
        if short and len(msg.splitlines()) < 2:
            print('', f'{header} {msg}', **kwargs)
            return
        print('', header, textwrap.indent(msg, '  '), **kwargs)

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # Check if the function is importable
        test_func = func.__name__
        assert globals()[test_func].__subproc_test_inner__ is func

        # Check if the arguments are round-trippable
        assert literal_eval(repr(args)) == args
        assert literal_eval(repr(kwargs)) == kwargs

        # Write a test script
        args_reprs = ([repr(a) for a in args]
                      + [f'{k}={v!r}' for k, v in kwargs.items()])
        if len(args_reprs) > 1:
            args_repr = '\n' + textwrap.indent(',\n'.join(args_reprs), ' ' * 8)
        else:
            args_repr = ', '.join(args_reprs)
        code_template = strip("""
        import sys
        sys.path.insert(0,  # Let the test import from this file
                        {path!r})
        from {mod} import (  # Import the test func from this file
            {test} as _{test})

        def {test}():
            _{test}.__subproc_test_inner__({args})

        if __name__ == '__main__':
            {test}()
        """)
        test_dir, test_filename = os.path.split(__file__)
        test_module_name, dot_py = os.path.splitext(test_filename)
        assert dot_py == '.py'
        code = code_template.format(path=test_dir, mod=test_module_name,
                                    test=test_func, args=args_repr)
        # Run the test script in a subprocess
        if debug:  # Use `pytest` to get perks like assertion rewriting
            cmd = [sys.executable, '-m', 'pytest', '-s']
        else:
            cmd = [sys.executable]
        with tempfile.TemporaryDirectory() as tmpdir:
            curdir = os.path.abspath(os.curdir)
            os.chdir(tmpdir)
            fname = 'my_test.py'
            cmd.append(fname)
            message(code, f'Test code run ({shlex.quote(fname)})')
            message(shlex.join(cmd), 'Executed command')
            try:
                with open(fname, mode='w') as fobj:
                    print(code, file=fobj)
                env = os.environ.copy()
                # Make sure that we're testing the "default behavior"
                env.pop('LINE_PROFILER_CORE', '')
                proc = subprocess.run(
                    cmd, capture_output=True, env=env, text=True)
            finally:
                os.chdir(curdir)
        if proc.stdout:
            message(proc.stdout, 'Stdout')
        else:
            message('<N/A>', 'Stdout', short=True)
        if proc.stderr:
            message(proc.stderr, 'Stderr', file=sys.stderr)
        else:
            message('<N/A>', 'Stderr', short=True)
        proc.check_returncode()

    wrapper.__subproc_test_inner__ = func
    return wrapper


def foo(n: int) -> int:
    result = 0
    for spam in range(1, n + 1):
        result += spam
    return result


def bar(n: int) -> int:
    result = 0
    for ham in range(1, n + 1):
        result += ham
    return result


def baz(n: int) -> int:
    result = 0
    for eggs in range(1, n + 1):
        result += eggs
    return result


class suspend_tracing:
    def __init__(self):
        self.callback = None
        self.events = 0

    def __enter__(self):
        if USE_SYS_MONITORING:
            mod = sys.monitoring
            self.events = mod.get_events(mod.PROFILER_ID)
            mod.set_events(mod.PROFILER_ID, mod.events.NO_EVENTS)
        else:
            self.callback = sys.gettrace()
            sys.settrace(None)

    def __exit__(self, *_, **__):
        if USE_SYS_MONITORING:
            mod = sys.monitoring
            mod.set_events(mod.PROFILER_ID, self.events)
            self.events = 0
        else:
            sys.settrace(self.callback)
            self.callback = None


def get_incr_logger(logs: List[str], func: Literal[foo, bar, baz] = foo, *,
                    bugged: bool = False,
                    report_return: bool = False) -> TracingFunc:
    '''
    Append a '<func>: spam = <...>' message whenever we hit the line in
    `func()` containing the incrementation of `result`.
    If it's made `bugged`, it sets the frame's `.f_trace_lines` to false
    after writing the first log entry, disabling line events.
    If `report_return` is true, a 'Returning from <func>()' log entry
    is written on return.
    '''
    def callback(
            frame: FrameType, event: Event, _) -> Union[TracingFunc, None]:
        if DEBUG and callback.emit_debug:
            print('{0.co_filename}:{1.f_lineno} - {0.co_name} ({2})'
                  .format(frame.f_code, frame, event))
        if event == 'call':  # Set up tracing for nested scopes
            return callback
        if event not in events:  # Only trace the specified events
            return
        code = frame.f_code
        if code.co_filename != filename or code.co_name != func_name:
            return
        if event == 'return':  # Write a return entry where appropriate
            logs.append(f'Returning from `{func_name}()`')
            return
        if frame.f_lineno == lineno:
            # Add log entry whenever the target line is hit
            counter_value = frame.f_locals.get(counter)
            logs.append(f'{func_name}: {counter} = {counter_value}')
            if bugged:  # Line-event tracing turned off after first hit
                frame.f_trace_lines = False
        return callback

    # Get data from `func()`: its (file-)name, the line number of the
    # incrementation, and the name of the counter variable
    func_name = func.__name__
    filename = func.__code__.co_filename
    lineno = func.__code__.co_firstlineno
    block = inspect.getblock(linecache.getlines(__file__)[lineno - 1:])
    (offset, line), = ((i, line) for i, line in enumerate(block)
                       if 'result +=' in line)
    lineno += offset
    counter = line.split()[-1]

    events = {'line'}
    if report_return:
        events.add('return')

    callback.emit_debug = False
    return callback


def get_return_logger(logs: List[str], *, bugged: bool = False) -> TracingFunc:
    '''
    Append a 'Returning from `<func>()`' message whenever we hit return
    from a function defined in this file. If it's made `bugged`, it
    panics and errors out when returning from `bar`, thus unsetting the
    `sys` trace.
    '''
    def callback(
            frame: FrameType, event: Event, _) -> Union[TracingFunc, None]:
        if DEBUG and callback.emit_debug:
            print('{0.co_filename}:{1.f_lineno} - {0.co_name} ({2})'
                  .format(frame.f_code, frame, event))
        if event == 'call':
            # Set up tracing for nested scopes
            return callback
        if event != 'return':
            return  # Only trace return events
        code = frame.f_code
        if code.co_filename != __file__:
            return  # Only trace functions in this file
        # Add log entry
        logs.append(f'Returning from `{code.co_name}()`')
        if bugged and code.co_name == 'bar':
            # Error out and cause `sys.settrace(None)`
            raise MyException

    callback.emit_debug = False
    return callback


class MyException(Exception):
    """Unique exception raised by some of the tests."""
    pass


# Tests


def _test_helper_callback_preservation(
        callback: Union[TracingFunc, None]) -> None:
    sys.settrace(callback)
    assert sys.gettrace() is callback, f'can\'t set trace to {callback!r}'
    profile = LineProfiler(wrap_trace=False)
    profile.enable_by_count()
    if not USE_SYS_MONITORING:
        assert profile in sys.gettrace().active_instances, (
            'can\'t set trace to the profiler')
    profile.disable_by_count()
    assert sys.gettrace() is callback, f'trace not restored to {callback!r}'
    sys.settrace(None)


@isolate_test_in_subproc
def test_callback_preservation():
    """
    Test in a subprocess that the profiler restores the active `sys`
    trace callback (or the lack thereof) after it's `.disable()`-ed.
    """
    _test_helper_callback_preservation(None)
    _test_helper_callback_preservation(lambda frame, event, arg: None)


@pytest.mark.parametrize('set_frame_local_trace', [True, False])
@pytest.mark.parametrize(
    ('label', 'use_profiler', 'wrap_trace'),
    [('base case', False, False),
     ('profiled (trace suspended)', True, False),
     ('profiled (trace wrapped)', True, True)])
@isolate_test_in_subproc
def test_callback_wrapping(
        label: str, use_profiler: bool,
        wrap_trace: bool, set_frame_local_trace: bool) -> None:
    """
    Test in a subprocess that the profiler can wrap around an existing
    trace callback such that we both profile the code and do whatever
    the existing callback does.
    """
    logs = []
    my_callback = get_incr_logger(logs)
    sys.settrace(my_callback)

    if use_profiler:
        profile = LineProfiler(
            wrap_trace=wrap_trace, set_frame_local_trace=set_frame_local_trace)
        foo_like = profile(foo)
        trace_preserved = wrap_trace
    else:
        foo_like = foo
        trace_preserved = True
    if trace_preserved or USE_SYS_MONITORING:
        exp_logs = [f'foo: spam = {spam}' for spam in range(1, 6)]
    else:
        exp_logs = []

    assert sys.gettrace() is my_callback, 'can\'t set custom trace'
    my_callback.emit_debug = True
    x = foo_like(5)
    my_callback.emit_debug = False
    assert x == 15, f'expected `foo(5) = 15`, got {x!r}'
    assert sys.gettrace() is my_callback, 'trace not restored afterwards'

    # Check that the existing trace function has been called where
    # appropriate
    print(f'Logs: {logs!r}')
    assert logs == exp_logs, f'expected logs = {exp_logs!r}, got {logs!r}'

    # Check that the profiling is as expected: 5 hits on the
    # incrementation line
    if not use_profiler:
        return
    with StringIO() as sio:
        profile.print_stats(stream=sio, summarize=True)
        out = sio.getvalue()
    print(out)
    line, = (line for line in out.splitlines() if '+=' in line)
    nhits = int(line.split()[1])
    assert nhits == 5, f'expected 5 profiler hits, got {nhits!r}'


@pytest.mark.parametrize(
    ('label', 'use_profiler', 'enable_count'),
    [('base case', False, 0),
     ('profiled (isolated)', True, 0),
     ('profiled (continuous)', True, 1)])
@isolate_test_in_subproc
def test_wrapping_throwing_callback(
        label: str, use_profiler: bool, enable_count: int) -> None:
    """
    Test in a subprocess that if the profiler wraps around an existing
    trace callback that errors out:
    - Profiling continues uninterrupted.
    - The errored-out trace callback is no longer called from the
      profiling traceback.
    - The `sys` traceback is set to `None` when the profiler is
      `.disable()`-ed.

    Notes
    -----
    Extra `enable_count` means that the profiler stays enabled between
    the calls to the profiled functions, and we thereyby test against
    these problematic behaviors after `my_callback()` bugs out:
    - If the profiler stops profiling (because the `sys` trace callback
      is unset), or
    - If the profiler's callback keeps calling `my_callback()`
      afterwards.
    """
    logs = []
    my_callback = get_return_logger(logs, bugged=True)
    sys.settrace(my_callback)
    assert sys.gettrace() is my_callback, 'can\'t set custom trace'

    if use_profiler:
        profile = LineProfiler(wrap_trace=True)
        foo_like, bar_like, baz_like = profile(foo), profile(bar), profile(baz)
    else:
        foo_like, bar_like, baz_like = foo, bar, baz
        enable_count = 0

    for _ in range(enable_count):
        profile.enable_by_count()
    my_callback.emit_debug = True
    x = foo_like(3)  # This is logged
    try:
        _ = bar_like(4)  # This is also logged, but...
    except MyException:
        # ... the trace func errors out as `bar()` returns, and as such
        # disables itself
        pass
    else:
        assert False, 'tracing function didn\'t error out'
    y = baz_like(5)  # Not logged because trace disabled itself
    my_callback.emit_debug = False
    for _ in range(enable_count):
        profile.disable_by_count()

    assert x == 6, f'expected `foo(3) = 6`, got {x!r}'
    assert y == 15, f'expected `baz(5) = 15`, got {y!r}'
    assert sys.gettrace() is None, (
        '`sys` trace = {sys.gettrace()!r} not reset afterwards')

    # Check that the existing trace function has been called where
    # appropriate
    print(f'Logs: {logs!r}')
    exp_logs = ['Returning from `foo()`', 'Returning from `bar()`']
    assert logs == exp_logs, f'expected logs = {exp_logs!r}, got {logs!r}'

    # Check that the profiling is as expected: 3 (resp. 4, 5) hits on
    # the incrementation line for `foo()` (resp. `bar()`, `baz()`)
    if not use_profiler:
        return
    with StringIO() as sio:
        profile.print_stats(stream=sio, summarize=True)
        out = sio.getvalue()
    print(out)
    for func, marker, exp_nhits in [('foo', 'spam', 3), ('bar', 'ham', 4),
                                    ('baz', 'eggs', 5)]:
        line, = (line for line in out.splitlines()
                 if line.endswith('+= ' + marker))
        nhits = int(line.split()[1])
        assert nhits == exp_nhits, (f'expected {exp_nhits} '
                                    f'profiler hits, got {nhits!r}')


@pytest.mark.parametrize(('label', 'use_profiler'),
                         [('base case', False), ('profiled', True)])
@isolate_test_in_subproc
def test_wrapping_line_event_disabling_callback(label: str,
                                                use_profiler: bool) -> None:
    """
    Test in a subprocess that if the profiler wraps around an existing
    trace callback that disables `.f_trace_lines`:
    - Profiling continues uninterrupted.
    - `.f_trace` is subsequently disabled, but only for line events in
      that frame.
    """
    logs = []
    my_callback = get_incr_logger(logs, bugged=True, report_return=True)
    sys.settrace(my_callback)

    if use_profiler:
        profile = LineProfiler(wrap_trace=True)
        foo_like = profile(foo)
    else:
        foo_like = foo

    assert sys.gettrace() is my_callback, 'can\'t set custom trace'
    my_callback.emit_debug = True
    x = foo_like(5)
    my_callback.emit_debug = False
    assert x == 15, f'expected `foo(5) = 15`, got {x!r}'
    assert sys.gettrace() is my_callback, 'trace not restored afterwards'

    # Check that the trace function has been called exactly once on the
    # line event, and once on the return event
    print(f'Logs: {logs!r}')
    exp_logs = ['foo: spam = 1', 'Returning from `foo()`']
    assert logs == exp_logs, f'expected logs = {exp_logs!r}, got {logs!r}'

    # Check that the profiling is as expected: 5 hits on the
    # incrementation line
    if not use_profiler:
        return
    with StringIO() as sio:
        profile.print_stats(stream=sio, summarize=True)
        out = sio.getvalue()
    print(out)
    line, = (line for line in out.splitlines() if '+=' in line)
    nhits = int(line.split()[1])
    assert nhits == 5, f'expected 5 profiler hits, got {nhits!r}'


def _test_helper_wrapping_thread_local_callbacks(
        profile: Union[LineProfiler, None], sleep: float = .0625) -> str:
    logs = []
    if threading.current_thread() == threading.main_thread():
        thread_label = 'main'
        func = foo
        my_callback = get_incr_logger(logs, func)
        exp_logs = [f'foo: spam = {spam}' for spam in range(1, 6)]
    else:
        thread_label = 'side'
        func = bar
        my_callback = get_return_logger(logs)
        exp_logs = ['Returning from `bar()`']
    if profile is None:
        func_like = func
    else:
        func_like = profile(func)
    print(f'Thread: {threading.get_ident()} ({thread_label})')

    # Check result
    sys.settrace(my_callback)
    assert sys.gettrace() is my_callback, 'can\'t set custom trace'
    my_callback.emit_debug = True
    x = func_like(5)
    my_callback.emit_debug = False
    assert x == 15, f'expected `{func.__name__}(5) = 15`, got {x!r}'
    assert sys.gettrace() is my_callback, 'trace not restored afterwards'

    # Check logs
    print(f'Logs ({thread_label} thread): {logs!r}')
    assert logs == exp_logs, f'expected logs = {exp_logs!r}, got {logs!r}'
    time.sleep(sleep)
    return '\n'.join(logs)


@pytest.mark.parametrize(('label', 'use_profiler'),
                         [('base case', False), ('profiled', True)])
@isolate_test_in_subproc
def test_wrapping_thread_local_callbacks(label: str,
                                         use_profiler: bool) -> None:
    """
    Test in a subprocess that the profiler properly handles thread-local
    `sys` trace callbacks.
    """
    profile = LineProfiler(wrap_trace=True) if use_profiler else None
    expected_results = {
        # From the main thread
        '\n'.join(f'foo: spam = {spam}' for spam in range(1, 6)),
        # From the other thread
        'Returning from `bar()`',
    }

    # Run tasks (and so some basic checks)
    results = set()
    with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
        tasks = []
        tasks.append(executor.submit(  # This is run on a side thread
            _test_helper_wrapping_thread_local_callbacks, profile))
        # This is run on the main thread
        results.add(_test_helper_wrapping_thread_local_callbacks(profile))
        results.update(
            future.result()
            for future in concurrent.futures.as_completed(tasks))
    assert results == expected_results, (f'expected {expected_results!r}, '
                                         f'got {results!r}')

    # Check profiling
    if profile is None:
        return
    with StringIO() as sio:
        profile.print_stats(stream=sio, summarize=True)
        out = sio.getvalue()
    print(out)
    for var in 'spam', 'ham':
        line, = (line for line in out.splitlines()
                 if line.endswith('+= ' + var))
        nhits = int(line.split()[1])
        assert nhits == 5, f'expected 5 profiler hits, got {nhits!r}'


@pytest.mark.parametrize(
    ('stay_in_scope', 'set_frame_local_trace', 'n', 'nhits'),
    [(True, True, 100, {0: 2,  # Both calls are traced
                        5: 0,  # Tracing suspended
                        7: 100}),  # Tracing restored (both calls)
     # If `set_frame_local_trace` is false:
     # - When using legacy tracing, tracing is suspended for the rest of
     #   the frame
     # - Else, tracing is unaffected
     (True, False, 100,
      {0: 2, 5: 0, 7: 100 if USE_SYS_MONITORING else 0}),
     # Calling a function always triggers `<trace>.__call__()`
     (False, True, 100, {0: 1,  # Only one of the calls is traced
                         2: 100}),  # 100 hits on the line in the loop
     (False, False, 100, {0: 1, 2: 100})])
@isolate_test_in_subproc
def test_python_level_trace_manipulation(
        stay_in_scope, set_frame_local_trace, n, nhits):
    """
    Test that:
    - When Python code retrieves the trace object set by `line_profiler`
      with `sys.gettrace()` and later restores it via `sys.settrace()`,
      it doesn't break anything, and
    - Resumption of line profiling in the same frame thereafter happens
      if and only if `set_frame_local_trace` is true.
    """
    prof = LineProfiler(set_frame_local_trace=set_frame_local_trace)

    def func_no_break(n):
        x = 0
        for n in range(1, n + 1):
            x += n
        return x

    @prof
    def func_break_in_middle(n):
        x = 0
        n_not_traced = n // 2
        n_traced = n - n_not_traced
        with suspend_tracing():
            for n in range(1, n_not_traced + 1):
                x += n
        for n in range(n_not_traced + 1, n_not_traced + n_traced + 1):
            x += n
        return x

    prof.add_callable(func_no_break)
    expected = n * (n + 1) // 2

    if stay_in_scope:
        # Do two calls, each with tracing suspended for half of the loop
        outer_ctx = inner_ctx = nullcontext()
        func = func_break_in_middle
    else:
        # Do one call with tracing suspended, then another with it
        # restored
        outer_ctx = prof
        inner_ctx = suspend_tracing()
        func = func_no_break
    with outer_ctx:
        with inner_ctx:
            assert func(n) == expected
        assert func(n) == expected
    timings = prof.get_stats().timings
    print(timings)
    prof.print_stats()
    entries = next(entries for (*_, func_name), entries in timings.items()
                   if func_name.endswith(func.__name__))
    body_start_line = min(lineno for (lineno, *_) in entries)
    all_nhits = {lineno - body_start_line: _nhits
                 for (lineno, _nhits, _) in entries}
    all_nhits = {lineno: all_nhits.get(lineno, 0) for lineno in nhits}
    assert all_nhits == nhits, f'expected {nhits=}, got {all_nhits=}'



import enum
try:
    from typing import Self  # type: ignore[attr-defined]  # noqa: F401
except ImportError:  # Python < 3.11
    from typing_extensions import Self  # noqa: F401


# Note: `mypy` tries to read this class as a free-standing enum
# (instead of an `enum.Enum` subclass that string enums are to inherit
# from), and complains that it has no members -- so silence that


class StringEnum(str, enum.Enum):  # type: ignore[misc]
    @staticmethod
    def _generate_next_value_(name: str, *_, **__) -> str:
        ...

    def __eq__(self, other) -> bool:
        ...

    def __str__(self) -> str:
        ...

    @classmethod
    def _missing_(cls, value) -> Self | None:
        ...



#!/usr/bin/env python
"""
Script to conveniently run profilers on code in a variety of
circumstances.

To profile a script, decorate the functions of interest with
:py:deco:`profile <line_profiler.explicit_profiler.GlobalProfiler>`:

.. code:: bash

    echo "if 1:
        @profile
        def main():
            1 + 1
        main()
    " > script_to_profile.py

NOTE:

    New in 4.1.0: Instead of relying on injecting :py:deco:`profile`
    into the builtins you can now ``import line_profiler`` and use
    :py:deco:`line_profiler.profile <line_profiler.explicit_profiler.GlobalProfiler>`
    to decorate your functions.  This allows the script to remain
    functional even if it is not actively profiled.  See
    :py:mod:`!line_profiler` (:ref:`link <line-profiler-basic-usage>`) for
    details.


Then run the script using :program:`kernprof`:

.. code:: bash

    kernprof -b script_to_profile.py

By default this runs with the default :py:mod:`cProfile` profiler and
does not require compiled modules. Instructions to view the results will
be given in the output. Alternatively, adding :option:`!-v` to the
command line will write results to stdout.

To enable line-by-line profiling, :py:mod:`line_profiler` must be
available and compiled, and the :option:`!-l` argument should be added to
the :program:`kernprof` invocation:

.. code:: bash

    kernprof -lb script_to_profile.py

NOTE:

    New in 4.3.0: More code execution options are added:

    * :command:`kernprof <options> -m some.module <args to module>`
      parallels :command:`python -m` and runs the provided module as
      :py:mod:`__main__`.
    * :command:`kernprof <options> -c "some code" <args to code>`
      parallels :command:`python -c` and executes the provided literal
      code.
    * :command:`kernprof <options> - <args to code>` parallels
      :command:`python -` and executes literal code passed via the
      :file:`stdin`.

    See also
    :doc:`kernprof invocations </manual/examples/example_kernprof>`.

For more details and options, refer to the CLI help.
To view the :program:`kernprof` help text run:

.. code:: bash

    kernprof --help

which displays:

.. code::

    usage: kernprof [-h] [-V] [--config CONFIG] [--no-config]
                    [--line-by-line [Y[es] | N[o] | T[rue] | F[alse] | on | off | 1 | 0]]
                    [--builtin [Y[es] | N[o] | T[rue] | F[alse] | on | off | 1 | 0]]
                    [-s SETUP] [-p {path/to/script | object.dotted.path}[,...]]
                    [--preimports [Y[es] | N[o] | T[rue] | F[alse] | on | off | 1 | 0]]
                    [--prof-imports [Y[es] | N[o] | T[rue] | F[alse] | on | off | 1 | 0]]
                    [-o OUTFILE] [-v] [-q]
                    [--rich [Y[es] | N[o] | T[rue] | F[alse] | on | off | 1 | 0]]
                    [-u UNIT]
                    [--skip-zero [Y[es] | N[o] | T[rue] | F[alse] | on | off | 1 | 0]]
                    [--summarize [Y[es] | N[o] | T[rue] | F[alse] | on | off | 1 | 0]]
                    [-i [OUTPUT_INTERVAL]]
                    {path/to/script | -m path.to.module | -c "literal code"} ...

    Run and profile a python script or module.

    positional arguments:
      {path/to/script | -m path.to.module | -c "literal code"}
                            The python script file, module, or literal code to run
      args                  Optional script arguments

    options:
      -h, --help            show this help message and exit
      -V, --version         show program's version number and exit
      --config CONFIG       Path to the TOML file, from the `tool.line_profiler.kernprof`
                            table of which to load defaults for the options. (Default:
                            'pyproject.toml')
      --no-config           Disable the loading of configuration files other than the
                            default one

    profiling options:
      --line-by-line [Y[es] | N[o] | T[rue] | F[alse] | on | off | 1 | 0]
                            Use the line-by-line profiler instead of cProfile. Implies
                            `--builtin`. (Default: False; short form: -l)
      --builtin [Y[es] | N[o] | T[rue] | F[alse] | on | off | 1 | 0]
                            Put `profile` in the builtins. Use
                            `profile.enable()`/`.disable()` to toggle profiling,
                            `@profile` to decorate functions, or `with profile:` to
                            profile a section of code. (Default: False; short form: -b)
      -s, --setup SETUP     Path to the Python source file containing setup code to
                            execute before the code to profile. (Default: N/A)
      -p, --prof-mod PROF_MOD
                            List of modules, functions and/or classes to profile specified
                            by their name or path. These profiling targets can be supplied
                            both as comma-separated items, or separately with multiple
                            copies of this flag. Packages are automatically recursed into
                            unless they are specified with `<pkg>.__init__`. Adding the
                            current script/module profiles the entirety of it. Only works
                            with line profiling (`-l`/`--line-by-line`). (Default: N/A;
                            pass an empty string to clear the defaults (or any `-p` target
                            specified earlier)
      ---preimports [Y[es] | N[o] | T[rue] | F[alse] | on | off | 1 | 0]
                            Instead of eagerly importing all profiling targets specified
                            via `-p` and profiling them, only profile those that are
                            directly imported in the profiled code. Only works with
                            line profiling (`-l`/`--line-by-line`). (Default: False)
                            Eagerly import all profiling targets specified via `-p` and
                            profile them, instead of only profiling those that are
                            directly imported in the profiled code. Only works with line
                            profiling (`-l`/`--line-by-line`). (Default: True)
      --prof-imports [Y[es] | N[o] | T[rue] | F[alse] | on | off | 1 | 0]
                            If the script/module profiled is in `--prof-mod`, autoprofile
                            all its imports. Only works with line profiling (`-l`/`--line-
                            by-line`). (Default: False)

    output options:
      -o, --outfile OUTFILE
                            Save stats to OUTFILE. (Default:
                            '<script_or_module_name>.lprof' in line-profiling mode
                            (`-l`/`--line-by-line`); '<script_or_module_name>.prof'
                            otherwise)
      -v, --verbose, --view
                            Increase verbosity level (default: 0). At level 1, view the
                            profiling results in addition to saving them; at level 2,
                            show other diagnostic info.
      -q, --quiet           Decrease verbosity level (default: 0). At level -1, disable
                            helpful messages (e.g. "Wrote profile results to <...>"); at
                            level -2, silence the stdout; at level -3, silence the stderr.
      --rich [Y[es] | N[o] | T[rue] | F[alse] | on | off | 1 | 0]
                            Use rich formatting if viewing output. (Default: False; short
                            form: -r)
      -u, --unit UNIT       Output unit (in seconds) in which the timing info is
                            displayed. (Default: 1e-06 s)
      --skip-zero [Y[es] | N[o] | T[rue] | F[alse] | on | off | 1 | 0]
                            Hide functions which have not been called. (Default: False;
                            short form: -z)
      --summarize [Y[es] | N[o] | T[rue] | F[alse] | on | off | 1 | 0]
                            Print a summary of total function time. (Default: False)
      -i, --output-interval [OUTPUT_INTERVAL]
                            Enables outputting of cumulative profiling results to OUTFILE
                            every OUTPUT_INTERVAL seconds. Uses the threading module.
                            Minimum value (and the value implied if the bare option is
                            given) is 1 s. (Default: 0 s (disabled))

NOTE:

    New in 5.0.0: For more intuitive profiling behavior, profiling
    targets in :option:`!--prof-mod` (except the profiled script/code)
    are now:

    * Eagerly pre-imported to be profiled (see
      :py:mod:`line_profiler.autoprofile.eager_preimports`),
      regardless of whether those imports directly occur in the profiled
      script/module/code.
    * Descended/Recursed into if they are packages; pass
      ``<pkg_name>.__init__`` instead of ``<pkg_name>`` to curtail
      descent and limit profiling to classes and functions in the local
      namespace of the :file:`__init__.py`.

    To restore the old behavior, pass the :option:`!--no-preimports`
    flag.
"""  # noqa: E501
import atexit
import builtins
import functools
import os
import sys
import threading
import asyncio  # NOQA
import concurrent.futures  # NOQA
import contextlib
import shutil
import tempfile
import time
import warnings
from argparse import ArgumentParser
from io import StringIO
from operator import methodcaller
from runpy import run_module
from pathlib import Path
from pprint import pformat
from shlex import quote
from textwrap import indent, dedent
from types import MethodType, SimpleNamespace

# NOTE: This version needs to be manually maintained in
# line_profiler/line_profiler.py and line_profiler/__init__.py as well
__version__ = '5.0.1'

# Guard the import of cProfile such that 3.x people
# without lsprof can still use this script.
try:
    from cProfile import Profile
except ImportError:
    from profile import Profile  # type: ignore[assignment,no-redef]

import line_profiler
from line_profiler.cli_utils import (
    add_argument, get_cli_config,
    get_python_executable as _python_command,  # Compatibility
    positive_float, short_string_path)
from line_profiler.profiler_mixin import ByCountProfilerMixin
from line_profiler._logger import Logger
from line_profiler import _diagnostics as diagnostics


DIAGNOSITICS_VERBOSITY = 2


def execfile(filename, globals=None, locals=None):
    """ Python 3.x doesn't have :py:func:`execfile` builtin """
    with open(filename, 'rb') as f:
        exec(compile(f.read(), filename, 'exec'), globals, locals)
# =====================================


class ContextualProfile(ByCountProfilerMixin, Profile):
    """ A subclass of :py:class:`Profile` that adds a context manager
    for Python 2.5 with: statements and a decorator.
    """
    def __init__(self, *args, **kwds):
        super(ByCountProfilerMixin, self).__init__(*args, **kwds)
        self.enable_count = 0

    def __call__(self, func):
        return self.wrap_callable(func)

    def enable_by_count(self, subcalls=True, builtins=True):
        """ Enable the profiler if it hasn't been enabled before.
        """
        if self.enable_count == 0:
            self.enable(subcalls=subcalls, builtins=builtins)
        self.enable_count += 1

    def disable_by_count(self):
        """ Disable the profiler if the number of disable requests matches the
        number of enable requests.
        """
        if self.enable_count > 0:
            self.enable_count -= 1
            if self.enable_count == 0:
                self.disable()

    # FIXME: `profile.Profile` is fundamentally incompatible with the
    # by-count paradigm we use, as it can't be `.enable()`-ed nor
    # `.disable()`-ed


class RepeatedTimer:
    """
    Background timer for outputting file every ``n`` seconds.

    Adapted from [SO474528]_.

    References:
        .. [SO474528] https://stackoverflow.com/questions/474528/execute-function-every-x-seconds/40965385#40965385
    """  # noqa: E501
    def __init__(self, interval, dump_func, outfile):
        self._timer = None
        self.interval = interval
        self.dump_func = dump_func
        self.outfile = outfile
        self.is_running = False
        self.next_call = time.time()
        self.start()

    def _run(self):
        self.is_running = False
        self.start()
        self.dump_func(self.outfile)

    def start(self):
        if not self.is_running:
            self.next_call += self.interval
            self._timer = threading.Timer(self.next_call - time.time(),
                                          self._run)
            self._timer.start()
            self.is_running = True

    def stop(self):
        self._timer.cancel()
        self.is_running = False


def find_module_script(module_name, *, static=True, exit_on_error=True):
    """Find the path to the executable script for a module or package."""
    from line_profiler.autoprofile.util_static import modname_to_modpath
    from importlib.util import find_spec

    def resolve_module_path(mod_name):  # type: (str) -> str | None
        try:
            mod_spec = find_spec(mod_name)
        except ImportError:
            return None
        if not mod_spec:
            return None
        fname = mod_spec.origin  # type: str | None
        if fname and os.path.exists(fname):
            return fname

    get_module_path = modname_to_modpath if static else resolve_module_path

    for suffix in '.__main__', '':
        fname = get_module_path(module_name + suffix)
        if fname:
            return fname

    msg = f'Could not find module `{module_name}`'
    if exit_on_error:
        print(msg, file=sys.stderr)
        raise SystemExit(1)
    else:
        raise ModuleNotFoundError(msg)


def find_script(script_name, *, exit_on_error=True):
    """ Find the script.

    If the input is not a file, then :envvar:`PATH` will be searched.
    """
    if os.path.isfile(script_name):
        return script_name
    path = os.getenv('PATH', os.defpath).split(os.pathsep)
    for dir in path:
        if dir == '':
            continue
        fn = os.path.join(dir, script_name)
        if os.path.isfile(fn):
            return fn

    msg = f'Could not find script {script_name!r}'
    if exit_on_error:
        print(msg, file=sys.stderr)
        raise SystemExit(1)
    else:
        raise FileNotFoundError(msg)


def _normalize_profiling_targets(targets):
    """
    Normalize the parsed :option:`!--prof-mod` by:

    * Normalizing file paths with :py:func:`find_script()`, and
      subsequently to absolute paths.
    * Splitting non-file paths at commas into (presumably) file paths
      and/or dotted paths.
    * Allowing paths specified earlier to be invalidated by an empty
      string.
    * Removing duplicates.
    """
    def find(path):
        try:
            path = find_script(path, exit_on_error=False)
        except FileNotFoundError:
            return None
        return os.path.abspath(path)

    results = {}
    for chunk in targets:
        if not chunk:
            results.clear()
            continue
        filename = find(chunk)
        if filename is not None:
            results.setdefault(filename)
            continue
        for subchunk in chunk.split(','):
            filename = find(subchunk)
            results.setdefault(subchunk if filename is None else filename)
    return list(results)


class _restore:
    """
    Restore a collection like :py:data:`sys.path` after running code
    which potentially modifies it.
    """
    def __init__(self, obj, getter, setter):
        self.obj = obj
        self.setter = setter
        self.getter = getter
        self.old = None

    def __enter__(self):
        assert self.old is None
        self.old = self.getter(self.obj)

    def __exit__(self, *_, **__):
        self.setter(self.obj, self.old)
        self.old = None

    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return wrapper

    @classmethod
    def sequence(cls, seq):
        """
        Example
        -------
        >>> l = [1, 2, 3]
        >>>
        >>> with _restore.sequence(l):
        ...     print(l)
        ...     l.append(4)
        ...     print(l)
        ...     l[:] = 5, 6
        ...     print(l)
        ...
        [1, 2, 3]
        [1, 2, 3, 4]
        [5, 6]
        >>> l
        [1, 2, 3]
        """
        def set_list(orig, copy):
            orig[:] = copy

        return cls(seq, methodcaller('copy'), set_list)

    @classmethod
    def mapping(cls, mpg):
        """
        Example
        -------
        >>> d = {1: 2}
        >>>
        >>> with _restore.mapping(d):
        ...     print(d)
        ...     d[2] = 3
        ...     print(d)
        ...     d.clear()
        ...     d.update({1: 4, 3: 5})
        ...     print(d)
        ...
        {1: 2}
        {1: 2, 2: 3}
        {1: 4, 3: 5}
        >>> d
        {1: 2}
        """
        def set_mapping(orig, copy):
            orig.clear()
            orig.update(copy)

        return cls(mpg, methodcaller('copy'), set_mapping)

    @classmethod
    def instance_dict(cls, obj):
        """
        Example
        -------
        >>> class Obj:
        ...     def __init__(self, x, y):
        ...         self.x, self.y = x, y
        ...
        ...     def __repr__(self):
        ...         return 'Obj({0.x!r}, {0.y!r})'.format(self)
        ...
        >>>
        >>> obj = Obj(1, 2)
        >>>
        >>> with _restore.instance_dict(obj):
        ...     print(obj)
        ...     obj.x, obj.y, obj.z = 4, 5, 6
        ...     print(obj, obj.z)
        ...
        Obj(1, 2)
        Obj(4, 5) 6
        >>> obj
        Obj(1, 2)
        >>> hasattr(obj, 'z')
        False
        """
        return cls.mapping(vars(obj))


def pre_parse_single_arg_directive(args, flag, sep='--'):
    """
    Pre-parse high-priority single-argument directives like
    :option:`!-m module` to emulate the behavior of
    :command:`python [...]`.

    Examples
    --------
    >>> import functools
    >>> pre_parse = functools.partial(pre_parse_single_arg_directive,
    ...                               flag='-m')

    Normal parsing:

    >>> pre_parse(['foo', 'bar', 'baz'])
    (['foo', 'bar', 'baz'], None, [])
    >>> pre_parse(['foo', 'bar', '-m', 'baz'])
    (['foo', 'bar'], 'baz', [])
    >>> pre_parse(['foo', 'bar', '-m', 'baz', 'foobar'])
    (['foo', 'bar'], 'baz', ['foobar'])

    Erroneous case:

    >>> pre_parse(['foo', 'bar', '-m'])
    Traceback (most recent call last):
      ...
    ValueError: argument expected for the -m option

    Prevent erroneous consumption of the flag by passing it `'--'`:

    >>> pre_parse(['foo', '--', 'bar', '-m', 'baz'])
    (['foo', '--'], None, ['bar', '-m', 'baz'])
    >>> pre_parse(['foo', '-m', 'spam',
    ...            'eggs', '--', 'bar', '-m', 'baz'])
    (['foo'], 'spam', ['eggs', '--', 'bar', '-m', 'baz'])
    """
    args = list(args)
    pre = []
    post = []
    try:
        i_sep = args.index(sep)
    except ValueError:  # No such element
        pass
    else:
        pre[:] = args[:i_sep]
        post[:] = args[i_sep + 1:]
        pre_pre, arg, pre_post = pre_parse_single_arg_directive(pre, flag)
        if arg is None:
            assert not pre_post
            return pre_pre + [sep], arg, post
        else:
            return pre_pre, arg, [*pre_post, sep, *post]
    try:
        i_flag = args.index(flag)
    except ValueError:  # No such element
        return args, None, []
    if i_flag == len(args) - 1:  # Last element
        raise ValueError(f'argument expected for the {flag} option')
    args, thing, post_args = args[:i_flag], args[i_flag + 1], args[i_flag + 2:]
    return args, thing, post_args


def no_op(*_, **__) -> None:
    pass


def _add_core_parser_arguments(parser):
    """
    Add the core kernprof args to a
    :py:class:`~argparse.ArgumentParser`.
    """
    default = get_cli_config('kernprof')
    add_argument(parser, '-V', '--version',
                 action='version', version=__version__)
    add_argument(parser, '--config',
                 help='Path to the TOML file, from the '
                 '`tool.line_profiler.kernprof` table of which to load '
                 'defaults for the options. '
                 f'(Default: {short_string_path(default.path)!r})')
    add_argument(parser, '--no-config',
                 action='store_const', dest='config', const=False,
                 help='Disable the loading of configuration files other '
                 'than the default one')
    prof_opts = parser.add_argument_group('profiling options')
    add_argument(prof_opts, '-l', '--line-by-line', action='store_true',
                 help='Use the line-by-line profiler instead of cProfile. '
                 'Implies `--builtin`. '
                 f'(Default: {default.conf_dict["line_by_line"]})')
    add_argument(prof_opts, '-b', '--builtin', action='store_true',
                 help="Put `profile` in the builtins. "
                 "Use `profile.enable()`/`.disable()` to "
                 "toggle profiling, "
                 "`@profile` to decorate functions, "
                 "or `with profile:` to profile a section of code. "
                 f"(Default: {default.conf_dict['builtin']})")
    if default.conf_dict['setup']:
        def_setupfile = repr(default.conf_dict['setup'])
    else:
        def_setupfile = 'N/A'
    add_argument(prof_opts, '-s', '--setup',
                 help='Path to the Python source file containing setup '
                 'code to execute before the code to profile. '
                 f'(Default: {def_setupfile})')
    if default.conf_dict['prof_mod']:
        def_prof_mod = repr(default.conf_dict['prof_mod'])
    else:
        def_prof_mod = 'N/A'
    add_argument(prof_opts, '-p', '--prof-mod', action='append',
                 help="List of modules, functions and/or classes to profile "
                 "specified by their name or path. These profiling targets "
                 "can be supplied both as comma-separated items, or "
                 "separately with multiple copies of this flag. Packages "
                 "are automatically recursed into unless they are specified "
                 "with `<pkg>.__init__`. Adding the current script/module "
                 "profiles the entirety of it. Only works with line "
                 "profiling (`-l`/`--line-by-line`). "
                 f"(Default: {def_prof_mod}; "
                 "pass an empty string to clear the defaults (or any `-p` "
                 "target specified earlier))")
    add_argument(prof_opts, '--preimports', action='store_true',
                 help="Eagerly import all profiling targets specified via "
                 "`-p` and profile them, instead of only profiling those "
                 "that are directly imported in the profiled code. "
                 "Only works with line profiling (`-l`/`--line-by-line`). "
                 f"(Default: {default.conf_dict['preimports']})")
    add_argument(prof_opts, '--prof-imports', action='store_true',
                 help="If the script/module profiled is in `--prof-mod`, "
                 "autoprofile all its imports. "
                 "Only works with line profiling (`-l`/`--line-by-line`). "
                 f"(Default: {default.conf_dict['prof_imports']})")
    out_opts = parser.add_argument_group('output options')
    if default.conf_dict['outfile']:
        def_outfile = repr(default.conf_dict['outfile'])
    else:
        def_outfile = (
            "'<script_or_module_name>.lprof' in line-profiling mode "
            "(`-l`/`--line-by-line`); "
            "'<script_or_module_name>.prof' otherwise")
    add_argument(out_opts, '-o', '--outfile',
                 help=f'Save stats to OUTFILE. (Default: {def_outfile})')
    add_argument(out_opts, '-v', '--verbose', '--view',
                 action='count', default=default.conf_dict['verbose'],
                 help="Increase verbosity level "
                 f"(default: {default.conf_dict['verbose']}). "
                 "At level 1, view the profiling results in addition to "
                 "saving them; "
                 "at level 2, show other diagnostic info.")
    add_argument(out_opts, '-q', '--quiet',
                 action='count', default=0,
                 help='Decrease verbosity level '
                 f"(default: {default.conf_dict['verbose']}). "
                 'At level -1, disable '
                 'helpful messages (e.g. "Wrote profile results to <...>"); '
                 'at level -2, silence the stdout; '
                 'at level -3, silence the stderr.')
    add_argument(out_opts, '-r', '--rich', action='store_true',
                 help='Use rich formatting if viewing output. '
                 f'(Default: {default.conf_dict["rich"]})')
    add_argument(out_opts, '-u', '--unit', type=positive_float,
                 help='Output unit (in seconds) in which '
                 'the timing info is displayed. '
                 f'(Default: {default.conf_dict["unit"]} s)')
    add_argument(out_opts, '-z', '--skip-zero', action='store_true',
                 help="Hide functions which have not been called. "
                 f"(Default: {default.conf_dict['skip_zero']})")
    add_argument(out_opts, '--summarize', action='store_true',
                 help='Print a summary of total function time. '
                 f'(Default: {default.conf_dict["summarize"]})')
    if default.conf_dict['output_interval']:
        def_out_int = f'{default.conf_dict["output_interval"]} s'
    else:
        def_out_int = '0 s (disabled)'
    add_argument(out_opts, '-i', '--output-interval',
                 type=int, const=1, nargs='?',
                 help="Enables outputting of cumulative profiling results "
                 "to OUTFILE every OUTPUT_INTERVAL seconds. "
                 "Uses the threading module. "
                 "Minimum value (and the value implied if the bare option "
                 f"is given) is 1 s. (Default: {def_out_int})")


def _build_parsers(args=None):
    parser_kwargs = {
        'description': 'Run and profile a python script.',
    }

    if args is None:
        args = sys.argv[1:]

    # Special cases: `kernprof [...] -m <module>` or
    # `kernprof [...] -c <script>` should terminate the parsing of all
    # subsequent options
    if '-m' in args and '-c' in args:
        special_mode = min(['-c', '-m'], key=args.index)
    elif '-m' in args:
        special_mode = '-m'
    else:
        special_mode = '-c'
    args, thing, post_args = pre_parse_single_arg_directive(args, special_mode)
    if special_mode == '-m':
        module, literal_code = thing, None
    else:
        module, literal_code = None, thing

    if module is literal_code is None:  # Normal execution
        real_parser, = parsers = [ArgumentParser(**parser_kwargs)]
        help_parser = None
    else:
        # We've already consumed the `-m <module>`, so we need a dummy
        # parser for generating the help text;
        # but the real parser should not consume the `options.script`
        # positional arg, and it it got the `--help` option, it should
        # hand off the the dummy parser
        real_parser = ArgumentParser(add_help=False, **parser_kwargs)
        real_parser.add_argument('-h', '--help', action='store_true')
        help_parser = ArgumentParser(**parser_kwargs)
        parsers = [real_parser, help_parser]
    for parser in parsers:
        _add_core_parser_arguments(parser)

        if parser is help_parser or module is literal_code is None:
            add_argument(parser, 'script',
                         metavar='{path/to/script'
                         ' | -m path.to.module | -c "literal code"}',
                         help='The python script file, module, or '
                         'literal code to run')
        add_argument(parser, 'args',
                     nargs='...', help='Optional script arguments')
    special_info = {
        'module': module,
        'literal_code': literal_code,
        'post_args': post_args,
        'args': args,
    }
    return real_parser, help_parser, special_info


def _parse_arguments(
        real_parser, help_parser, special_info, args, exit_on_error):

    module = special_info['module']
    literal_code = special_info['literal_code']
    post_args = special_info['post_args']

    # Hand off to the dummy parser if necessary to generate the help
    # text
    try:
        options = SimpleNamespace(**vars(real_parser.parse_args(args)))
    except SystemExit as e:
        # If `exit_on_error` is true, let `SystemExit` bubble up and
        # kill the interpretor;
        # else, catch and handle it more gracefully
        # (Note: can't use `ArgumentParser(exit_on_error=False)` in
        # Python 3.8)
        if exit_on_error:
            raise
        elif e.code:
            raise RuntimeError from None
        else:
            return
    # TODO: make flags later where appropriate
    options.dryrun = diagnostics.NO_EXEC
    options.static = diagnostics.STATIC_ANALYSIS
    if help_parser and getattr(options, 'help', False):
        help_parser.print_help()
        if exit_on_error:
            raise SystemExit(0)
        else:
            return

    # Parse the provided config file (if any), and resolve the values
    # of the un-specified options
    try:
        del options.help
    except AttributeError:
        pass
    default = get_cli_config('kernprof', options.config)
    options.config = default.path
    for key, default in default.conf_dict.items():
        if getattr(options, key, None) is None:
            setattr(options, key, default)

    # Add in the pre-partitioned arguments cut off by `-m <module>` or
    # `-c <script>`
    options.args += post_args
    if module is not None:
        options.script = module

    tempfile_source_and_content = None
    if literal_code is not None:
        tempfile_source_and_content = 'command', literal_code
    elif options.script == '-' and not module:
        tempfile_source_and_content = 'stdin', sys.stdin.read()

    # Handle output
    options.verbose -= options.quiet
    options.debug = (diagnostics.DEBUG
                     or options.verbose >= DIAGNOSITICS_VERBOSITY)
    logger_kwargs = {'name': 'kernprof'}
    logger_kwargs['backend'] = 'auto'
    if options.debug:
        # Debugging forces the stdlib logger
        logger_kwargs['verbose'] = 2
        logger_kwargs['backend'] = 'stdlib'
    elif options.verbose > -1:
        logger_kwargs['verbose'] = 1
    else:
        logger_kwargs['verbose'] = 0
    logger_kwargs['stream'] = {
        'format': '[%(name)s %(asctime)s %(levelname)s] %(message)s',
    }
    # Reinitialize the diagnostic logs, we are very likely the main script.
    diagnostics.log = Logger(**logger_kwargs)
    if options.rich:
        try:
            import rich  # noqa: F401
        except ImportError:
            options.rich = False
            diagnostics.log.debug('`rich` not installed, unsetting --rich')

    diagnostics.log.debug(
        f'Loaded configs from {short_string_path(options.config)!r}')
    return options, tempfile_source_and_content


@_restore.sequence(sys.argv)
@_restore.sequence(sys.path)
@_restore.instance_dict(diagnostics)
def main(args=None, *, exit_on_error=True):
    """
    Runs the command line interface

    Note:
        To help with traceback formatting, the deletion of temporary
        files created during execution may be deferred to when the
        interpreter exits.
    """
    real_parser, help_parser, special_info = _build_parsers(args=args)
    args = special_info['args']
    module = special_info['module']

    options, tempfile_source_and_content = _parse_arguments(
        real_parser, help_parser, special_info, args, exit_on_error)

    if module is not None:
        diagnostics.log.debug(f'Profiling module: {module}')
    elif tempfile_source_and_content:
        diagnostics.log.debug(
            f'Profiling script read from: {tempfile_source_and_content[0]}')
    else:
        diagnostics.log.debug(
            f'Profiling script: {short_string_path(options.script)!r}')

    with contextlib.ExitStack() as stack:
        enter = stack.enter_context
        if options.verbose < -1:  # Suppress stdout
            devnull = enter(open(os.devnull, mode='w'))
            enter(contextlib.redirect_stdout(devnull))
        if options.verbose < -2:  # Suppress stderr
            enter(contextlib.redirect_stderr(devnull))
        # Instead of relying on `tempfile.TemporaryDirectory`, manually
        # manage a tempdir to ensure that files exist at
        # traceback-formatting time if needs be
        options.tmpdir = tmpdir = tempfile.mkdtemp()
        if diagnostics.KEEP_TEMPDIRS:
            cleanup = no_op
        else:
            cleanup = functools.partial(
                _remove, tmpdir, recursive=True, missing_ok=True,
            )
        if tempfile_source_and_content:
            try:
                _write_tempfile(*tempfile_source_and_content, options)
            except Exception:
                # Tempfile creation failed, delete the tempdir ASAP
                cleanup()
                raise
        try:
            _main_profile(options, module, exit_on_error)
        except BaseException:
            # Defer deletion to after the traceback has been formatted
            # if needs be
            if os.listdir(tmpdir):
                atexit.register(cleanup)
            else:  # Empty tempdir, just delete it
                cleanup()
            raise
        else:  # Execution succeeded, delete the tempdir ASAP
            cleanup()


def _touch_tempfile(*args, **kwargs):
    """
    Wrapper around :py:func:`tempfile.mkstemp()` which drops and closes
    the integer handle (which we don't need and may cause issues on some
    platforms).
    """
    handle, path = tempfile.mkstemp(*args, **kwargs)
    try:
        os.close(handle)
    except Exception:
        os.remove(path)
        raise
    return path


def _write_tempfile(source, content, options):
    """
    Called by :py:func:`main()` to handle :command:`kernprof -c` and
    :command:`kernprof -`;
    not to be invoked on its own.
    """
    # Set up the script to be run
    file_prefix = f'kernprof-{source}'
    # Do what 3.14 does (#103998)... and also just to be user-friendly
    content = dedent(content)
    fname = os.path.join(options.tmpdir, file_prefix + '.py')
    with open(fname, mode='w') as fobj:
        print(content, file=fobj)
    diagnostics.log.debug(
        f'Wrote temporary script file to {short_string_path(fname)!r}:')
    options.script = fname
    # Add the tempfile to `--prof-mod`
    if options.prof_mod:
        options.prof_mod.append(fname)
    else:
        options.prof_mod = [fname]
    # Set the output file to somewhere nicer (also take care of possible
    # filename clash)
    if not options.outfile:
        extension = 'lprof' if options.line_by_line else 'prof'
        options.outfile = _touch_tempfile(dir=os.curdir,
                                          prefix=file_prefix + '-',
                                          suffix='.' + extension)
        diagnostics.log.debug(
            'Using default output destination '
            f'{short_string_path(options.outfile)!r}')


def _gather_preimport_targets(options, exclude):
    """
    Used in _write_preimports
    """
    from line_profiler.autoprofile.util_static import modpath_to_modname
    from line_profiler.autoprofile.eager_preimports import is_dotted_path
    filtered_targets = []
    recurse_targets = []
    invalid_targets = []
    for target in options.prof_mod:
        if is_dotted_path(target):
            modname = target
        else:
            # Paths already normalized by
            # `_normalize_profiling_targets()`
            if not os.path.exists(target):
                invalid_targets.append(target)
                continue
            if any(os.path.samefile(target, excluded) for excluded in exclude):
                # Ignore the script to be run in eager importing
                # (`line_profiler.autoprofile.autoprofile.run()` will
                # handle it)
                continue
            modname = modpath_to_modname(target, hide_init=False)
        if modname is None:  # Not import-able
            invalid_targets.append(target)
            continue
        if modname.endswith('.__init__'):
            modname = modname.rpartition('.')[0]
            filtered_targets.append(modname)
        else:
            recurse_targets.append(modname)
    if invalid_targets:
        invalid_targets = sorted(set(invalid_targets))
        msg = ('{} profile-on-import target{} cannot be converted to '
               'dotted-path form: {!r}'
               .format(len(invalid_targets),
                       '' if len(invalid_targets) == 1 else 's',
                       invalid_targets))
        warnings.warn(msg)
        diagnostics.log.warning(msg)

    return filtered_targets, recurse_targets


def _write_preimports(prof, options, exclude):
    """
    Called by :py:func:`main()` to handle eager pre-imports;
    not to be invoked on its own.
    """
    from line_profiler.autoprofile.eager_preimports import write_eager_import_module
    from line_profiler.autoprofile.autoprofile import (
        _extend_line_profiler_for_profiling_imports as upgrade_profiler)

    filtered_targets, recurse_targets = _gather_preimport_targets(options, exclude)
    if not (filtered_targets or recurse_targets):
        return
    # We could've done everything in-memory with `io.StringIO` and `exec()`,
    # but that results in indecipherable tracebacks should anything goes wrong;
    # so we write to a tempfile and `execfile()` it
    upgrade_profiler(prof)
    temp_mod_path = _touch_tempfile(dir=options.tmpdir,
                                    prefix='kernprof-eager-preimports-',
                                    suffix='.py')
    write_module_kwargs = {
        'dotted_paths': filtered_targets,
        'recurse': recurse_targets,
        'static': options.static,
    }
    temp_file = open(temp_mod_path, mode='w')
    if options.debug:
        with StringIO() as sio:
            write_eager_import_module(stream=sio, **write_module_kwargs)
            code = sio.getvalue()
        with temp_file as fobj:
            print(code, file=fobj)
        diagnostics.log.debug(
            'Wrote temporary module for pre-imports to '
            f'{short_string_path(temp_mod_path)!r}')
    else:
        with temp_file as fobj:
            write_eager_import_module(stream=fobj, **write_module_kwargs)
    if not options.dryrun:
        ns = {}  # Use a fresh namespace
        execfile(temp_mod_path, ns, ns)
    # Delete the tempfile ASAP if its execution succeeded
    if not diagnostics.KEEP_TEMPDIRS:
        _remove(temp_mod_path)


def _remove(path, *, recursive=False, missing_ok=False):
    path = Path(path)
    if path.is_dir():
        if recursive:
            shutil.rmtree(path, ignore_errors=missing_ok)
        else:
            path.rmdir()
    else:
        path.unlink(missing_ok=missing_ok)


def _dump_filtered_stats(tmpdir, prof, filename):
    import os
    import pickle

    # Build list of known temp file paths
    tempfile_paths = [
        os.path.join(dirpath, fname)
        for dirpath, _, fnames in os.walk(tmpdir)
        for fname in fnames
    ]

    if not tempfile_paths or isinstance(prof, ContextualProfile):
        # - No tempfiles written -> no function lives in tempfiles
        #   -> no need to filter anything
        # - Not using `line_profiler`
        #   -> doesn't matter if the source lines can't be retrieved
        #   -> no need to filter anything
        prof.dump_stats(filename)
        return

    # Filter the filenames to remove data from tempfiles, which will
    # have been deleted by the time the results are viewed in a
    # separate process
    stats = prof.get_stats()
    timings = stats.timings
    for key in set(timings):
        fname = key[0]
        try:
            if any(os.path.samefile(fname, tmp) for tmp in tempfile_paths):
                del timings[key]
        except OSError:
            del timings[key]

    with open(filename, 'wb') as f:
        pickle.dump(stats, f, protocol=pickle.HIGHEST_PROTOCOL)


def _format_call_message(func, *args, **kwargs):
    if isinstance(func, functools.partial):
        return _format_call_message(
            func.func, *func.args, *args, **{**func.keywords, **kwargs})
    if isinstance(func, MethodType):
        obj = func.__self__
        func_repr = (
            '{0.__module__}.{0.__qualname__}(...).{1.__name__}'
            .format(type(obj), func.__func__))
    else:
        try:
            func_repr = '{0.__module__}.{0.__qualname__}'.format(func)
        except Exception:  # Fallback
            func_repr = repr(func)
    args_repr = dedent(' ' + pformat(args)[len('['):-len(']')])
    lprefix = len('namespace(')
    kwargs_repr = dedent(
        ' ' * lprefix
        + pformat(SimpleNamespace(**kwargs))[lprefix:-len(')')])
    if args_repr and kwargs_repr:
        all_args_repr = f'{args_repr},\n{kwargs_repr}'
    else:
        all_args_repr = args_repr or kwargs_repr
    if all_args_repr:
        call = '{}(\n{})'.format(
            func_repr, indent(all_args_repr, '    '))
    else:
        call = func_repr + '()'
    return call


def _call_with_diagnostics(options, func, *args, **kwargs):
    if options.debug:
        call = _format_call_message(func, *args, **kwargs)
        diagnostics.log.debug(f'Calling: {call}')
    if options.dryrun:
        return
    return func(*args, **kwargs)


def _pre_profile(options, module, exit_on_error):
    """
    Prepare the environment to execute profiling with requested options.

    Note:
        modifies ``options`` with extra attributes.
    """
    if not options.outfile:
        extension = 'lprof' if options.line_by_line else 'prof'
        options.outfile = f'{os.path.basename(options.script)}.{extension}'
        diagnostics.log.debug(
            'Using default output destination '
            f'{short_string_path(options.outfile)!r}')

    sys.argv = [options.script] + options.args
    if module:
        # Make sure the current directory is on `sys.path` to emulate
        # `python -m`
        # Note: this NEEDS to happen here, before the setup script (or
        # any other code) has a chance to `os.chdir()`
        sys.path.insert(0, os.path.abspath(os.curdir))
    if options.setup:
        # Run some setup code outside of the profiler. This is good for large
        # imports.
        setup_file = find_script(options.setup, exit_on_error=exit_on_error)
        # Make sure the script's directory is on sys.path instead of just
        # kernprof.py's.
        sys.path.insert(0, os.path.dirname(setup_file))
        ns = {'__file__': setup_file, '__name__': '__main__'}
        diagnostics.log.debug(
            f'Executing file {short_string_path(setup_file)!r} '
            'as pre-profiling setup')
        if not options.dryrun:
            execfile(setup_file, ns, ns)

    if options.line_by_line:
        prof = line_profiler.LineProfiler()
        options.builtin = True
    elif Profile.__module__ == 'profile':
        raise RuntimeError('non-line-by-line profiling depends on cProfile, '
                           'which is not available on this platform')
    else:
        prof = ContextualProfile()

    # Overwrite the explicit decorator
    global_profiler = line_profiler.profile
    install_profiler = global_profiler._kernprof_overwrite
    install_profiler(prof)

    if options.builtin:
        builtins.__dict__['profile'] = prof

    if module:
        script_file = find_module_script(
            options.script, static=options.static, exit_on_error=exit_on_error)
    else:
        script_file = find_script(options.script, exit_on_error=exit_on_error)
        # Make sure the script's directory is on sys.path instead of
        # just kernprof.py's.
        sys.path.insert(0, os.path.dirname(script_file))

    # If using eager pre-imports, write a dummy module which contains
    # all those imports and marks them for profiling, then run it
    if options.prof_mod:
        # Note: `prof_mod` entries can be filenames (which can contain
        # commas), so check against existing filenames before splitting
        # them
        options.prof_mod = _normalize_profiling_targets(options.prof_mod)
    if not options.prof_mod:
        options.preimports = False
    if options.line_by_line and options.preimports:
        # We assume most items in `.prof_mod` to be import-able without
        # significant side effects, but the same cannot be said if it
        # contains the script file to be run. E.g. the script may not
        # even have a `if __name__ == '__main__': ...` guard. So don't
        # eager-import it.
        exclude = set() if module else {script_file}
        _write_preimports(prof, options, exclude)

    options.global_profiler = global_profiler
    options.install_profiler = install_profiler
    if options.output_interval and not options.dryrun:
        options.rt = RepeatedTimer(max(options.output_interval, 1),
                                   prof.dump_stats, options.outfile)
    else:
        options.rt = None
    options.original_stdout = sys.stdout
    return script_file, prof


def _main_profile(options, module=False, exit_on_error=True):
    """
    Called by :py:func:`main()` for the actual execution and profiling of code
    after initial parsing of options; not to be invoked on its own.
    """
    script_file, prof = _pre_profile(options, module, exit_on_error)
    try:
        rmod = functools.partial(run_module,
                                 run_name='__main__', alter_sys=True)
        ns = {'__file__': script_file, '__name__': '__main__',
              'execfile': execfile, 'rmod': rmod,
              'prof': prof}
        if options.prof_mod and options.line_by_line:
            from line_profiler.autoprofile import autoprofile
            _call_with_diagnostics(
                options,
                autoprofile.run, script_file, ns,
                prof_mod=options.prof_mod,
                profile_imports=options.prof_imports,
                as_module=module is not None)
        else:
            if module:
                runner, target = 'rmod', options.script
            else:
                runner, target = 'execfile', script_file
            assert runner in ns
            if options.builtin:
                _call_with_diagnostics(options, ns[runner], target, ns)
            else:
                _call_with_diagnostics(
                    options, prof.runctx,
                    f'{runner}({target!r}, globals())', ns, ns)
    finally:
        _post_profile(options, prof)


def _post_profile(options, prof):
    """
    Cleanup setup after executing a main profile
    """
    if options.rt is not None:
        options.rt.stop()
    if not options.dryrun:
        _dump_filtered_stats(options.tmpdir, prof, options.outfile)
    short_outfile = short_string_path(options.outfile)
    diagnostics.log.info(
        ('Profile results would have been written to '
         if options.dryrun else
         'Wrote profile results ')
        + f'to {short_outfile!r}')
    if options.verbose > 0 and not options.dryrun:
        kwargs = {}
        if not isinstance(prof, ContextualProfile):
            kwargs.update(output_unit=options.unit,
                          stripzeros=options.skip_zero,
                          summarize=options.summarize,
                          rich=options.rich,
                          stream=options.original_stdout,
                          config=options.config)
        _call_with_diagnostics(options, prof.print_stats, **kwargs)
    else:
        py_exe = _python_command()
        if isinstance(prof, ContextualProfile):
            show_mod = 'pstats'
        else:
            show_mod = 'line_profiler -rmt'
        diagnostics.log.info('Inspect results with:\n'
                             f'{quote(py_exe)} -m {show_mod} '
                             f'{quote(short_outfile)}')
    # Fully disable the profiler
    for _ in range(prof.enable_count):
        prof.disable_by_count()
    # Restore the state of the global `@line_profiler.profile`
    if options.global_profiler:
        options.install_profiler(None)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
This module defines the core :class:`LineProfiler` class as well as methods to
inspect its output. This depends on the :py:mod:`line_profiler._line_profiler`
Cython backend.
"""
import functools
import inspect
import linecache
import operator
import os
import pickle
import sys
import tempfile
import types
import tokenize
from argparse import ArgumentParser
from datetime import datetime

try:
    from ._line_profiler import (LineProfiler as CLineProfiler,
                                 LineStats as CLineStats)
except ImportError as ex:
    raise ImportError(
        'The line_profiler._line_profiler c-extension is not importable. '
        f'Has it been compiled? Underlying error is ex={ex!r}'
    )
from . import _diagnostics as diagnostics
from .cli_utils import (
    add_argument, get_cli_config, positive_float, short_string_path)
from .profiler_mixin import ByCountProfilerMixin, is_c_level_callable
from .scoping_policy import ScopingPolicy
from .toml_config import ConfigSource


# NOTE: This needs to be in sync with ../kernprof.py and __init__.py
__version__ = '5.0.1'


@functools.lru_cache()
def get_column_widths(config=False):
    """
    Arguments
        config (bool | str | pathlib.PurePath | None)
            Passed to :py:meth:`.ConfigSource.from_config`.
    Note:
        * Results are cached.
        * The default value (:py:data:`False`) loads the config from the
          default TOML file that the package ships with.
    """
    subconf = (ConfigSource.from_config(config)
               .get_subconfig('show', 'column_widths'))
    return types.MappingProxyType(subconf.conf_dict)


def load_ipython_extension(ip):
    """ API for IPython to recognize this module as an IPython extension.
    """
    from .ipython_extension import LineProfilerMagics
    ip.register_magics(LineProfilerMagics)


def get_code_block(filename, lineno):
    """
    Get the lines in the code block in a file starting from required
    line number; understands Cython code.

    Args:
        filename (Union[os.PathLike, str])
            Path to the source file.
        lineno (int)
            1-indexed line number of the first line in the block.

    Returns:
        lines (list[str])
            Newline-terminated string lines.

    Note:
        This function makes use of :py:func:`inspect.getblock`, which is
        public but undocumented API.  That said, it has been in use in
        this repo since 2008 (`fb60664`_), so we will continue using it
        until we can't.

        .. _fb60664: https://github.com/pyutils/line_profiler/commit/\
fb60664135296ba6061cfaa2bb66d4ba77964c53


    Example:
        >>> from os.path import join
        >>> from tempfile import TemporaryDirectory
        >>> from textwrap import dedent
        >>>
        >>>
        >>> def get_last_line(*args, **kwargs):
        ...     lines = get_code_block(*args, **kwargs)
        ...     return lines[-1].rstrip('\\n')
        ...
        >>>
        >>> with TemporaryDirectory() as tmpdir:
        ...     fname = join(tmpdir, 'cython_source.pyx')
        ...     with open(fname, mode='w') as fobj:
        ...         print(dedent('''
        ...     class NormalClass:                   # 1
        ...         def __init__(self):              # 2
        ...             pass                         # 3
        ...
        ...         def normal_method(self, *args):  # 5
        ...             pass                         # 6
        ...
        ...     cdef class CythonClass:              # 8
        ...         cpdef cython_method(self):       # 9
        ...             pass                         # 10
        ...
        ...         property legacy_cython_prop:     # 12
        ...             def __get__(self):           # 13
        ...                 return None              # 14
        ...             def __set__(self, value):    # 15
        ...                 pass                     # 16
        ...
        ...     def normal_func(x, y, z):            # 18
        ...         with some_ctx():                 # 19
        ...             ...                          # 20
        ...
        ...     cdef cython_function(                # 22
        ...             int x, int y, int z):        # 23
        ...         ...                              # 24
        ...                      ''').strip('\\n'),
        ...               file=fobj)
        ...     # Vanilla Python code blocks:
        ...     # - `NormalClass`
        ...     assert get_last_line(fname, 1).endswith('# 6')
        ...     # - `NormalClass.__init__()`
        ...     assert get_last_line(fname, 2).endswith('# 3')
        ...     # - `normal_func()`
        ...     assert get_last_line(fname, 18).endswith('# 20')
        ...     # Cython code blocks:
        ...     # - `CythonClass`
        ...     assert get_last_line(fname, 8).endswith('# 16')
        ...     # - `CythonClass.cython_method()`
        ...     assert get_last_line(fname, 9).endswith('# 10')
        ...     # - `CythonClass.legacy_cython_prop`
        ...     assert get_last_line(fname, 12).endswith('# 16')
        ...     # - `cython_function()`
        ...     assert get_last_line(fname, 22).endswith('# 24')
    """
    BlockFinder = inspect.BlockFinder
    namespace = inspect.getblock.__globals__
    namespace['BlockFinder'] = _CythonBlockFinder
    try:
        return inspect.getblock(linecache.getlines(filename)[lineno - 1:])
    finally:
        namespace['BlockFinder'] = BlockFinder


class _CythonBlockFinder(inspect.BlockFinder):
    """
    Compatibility layer turning Cython-specific code blocks (``cdef``,
    ``cpdef``, and legacy ``property`` declaration) into something that
    is understood by :py:class:`inspect.BlockFinder`.

    Note:
        This function makes use of :py:func:`inspect.BlockFinder`, which
        is public but undocumented API.  See similar caveat in
        :py:func:`~.get_code_block`.
    """
    def tokeneater(self, type, token, *args, **kwargs):
        if (
                not self.started
                and type == tokenize.NAME
                and token in ('cdef', 'cpdef', 'property')):
            # Fudge the token to get the desired 'scoping' behavior
            token = 'def'
        return super().tokeneater(type, token, *args, **kwargs)


class _WrapperInfo:
    """
    Helper object for holding the state of a wrapper function.

    Attributes:
        func (types.FunctionType):
            The function it wraps.
        profiler_id (int)
            ID of the `LineProfiler`.
    """
    def __init__(self, func, profiler_id):
        self.func = func
        self.profiler_id = profiler_id


class LineStats(CLineStats):
    def __repr__(self):
        return '{}({}, {:.2G})'.format(
            type(self).__name__, self.timings, self.unit)

    def __eq__(self, other):
        """
        Example:
            >>> from copy import deepcopy
            >>>
            >>>
            >>> stats1 = LineStats(
            ...     {('foo', 1, 'spam.py'): [(2, 10, 300)],
            ...      ('bar', 10, 'spam.py'):
            ...      [(11, 2, 1000), (12, 1, 500)]},
            ...     1E-6)
            >>> stats2 = deepcopy(stats1)
            >>> assert stats1 == stats2 is not stats1
            >>> stats2.timings = 1E-7
            >>> assert stats2 != stats1
            >>> stats3 = deepcopy(stats1)
            >>> assert stats1 == stats3 is not stats1
            >>> stats3.timings['foo', 1, 'spam.py'][:] = [(2, 11, 330)]
            >>> assert stats3 != stats1
        """
        for attr in 'timings', 'unit':
            getter = operator.attrgetter(attr)
            try:
                if getter(self) != getter(other):
                    return False
            except (AttributeError, TypeError):
                return NotImplemented
        return True

    def __add__(self, other):
        """
        Example:
            >>> stats1 = LineStats(
            ...     {('foo', 1, 'spam.py'): [(2, 10, 300)],
            ...      ('bar', 10, 'spam.py'):
            ...      [(11, 2, 1000), (12, 1, 500)]},
            ...     1E-6)
            >>> stats2 = LineStats(
            ...     {('bar', 10, 'spam.py'):
            ...      [(11, 10, 20000), (12, 5, 1000)],
            ...      ('baz', 5, 'eggs.py'): [(5, 2, 5000)]},
            ...     1E-7)
            >>> stats_sum = LineStats(
            ...     {('foo', 1, 'spam.py'): [(2, 10, 300)],
            ...      ('bar', 10, 'spam.py'):
            ...      [(11, 12, 3000), (12, 6, 600)],
            ...      ('baz', 5, 'eggs.py'): [(5, 2, 500)]},
            ...     1E-6)
            >>> assert stats1 + stats2 == stats2 + stats1 == stats_sum
        """
        timings, unit = self._get_aggregated_timings([self, other])
        return type(self)(timings, unit)

    def __iadd__(self, other):
        """
        Example:
            >>> stats1 = LineStats(
            ...     {('foo', 1, 'spam.py'): [(2, 10, 300)],
            ...      ('bar', 10, 'spam.py'):
            ...      [(11, 2, 1000), (12, 1, 500)]},
            ...     1E-6)
            >>> stats2 = LineStats(
            ...     {('bar', 10, 'spam.py'):
            ...      [(11, 10, 20000), (12, 5, 1000)],
            ...      ('baz', 5, 'eggs.py'): [(5, 2, 5000)]},
            ...     1E-7)
            >>> stats_sum = LineStats(
            ...     {('foo', 1, 'spam.py'): [(2, 10, 300)],
            ...      ('bar', 10, 'spam.py'):
            ...      [(11, 12, 3000), (12, 6, 600)],
            ...      ('baz', 5, 'eggs.py'): [(5, 2, 500)]},
            ...     1E-6)
            >>> address = id(stats2)
            >>> stats2 += stats1
            >>> assert id(stats2) == address
            >>> assert stats2 == stats_sum
        """
        self.timings, self.unit = self._get_aggregated_timings([self, other])
        return self

    def print(self, stream=None, **kwargs):
        show_text(self.timings, self.unit, stream=stream, **kwargs)

    def to_file(self, filename):
        """ Pickle the instance to the given filename.
        """
        with open(filename, 'wb') as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)

    @classmethod
    def from_files(cls, file, /, *files):
        """
        Utility function to load an instance from the given filenames.
        """
        stats_objs = []
        for file in [file, *files]:
            with open(file, 'rb') as f:
                stats_objs.append(pickle.load(f))
        return cls.from_stats_objects(*stats_objs)

    @classmethod
    def from_stats_objects(cls, stats, /, *more_stats):
        """
        Example:
            >>> stats1 = LineStats(
            ...     {('foo', 1, 'spam.py'): [(2, 10, 300)],
            ...      ('bar', 10, 'spam.py'):
            ...      [(11, 2, 1000), (12, 1, 500)]},
            ...     1E-6)
            >>> stats2 = LineStats(
            ...     {('bar', 10, 'spam.py'):
            ...      [(11, 10, 20000), (12, 5, 1000)],
            ...      ('baz', 5, 'eggs.py'): [(5, 2, 5000)]},
            ...     1E-7)
            >>> stats_combined = LineStats.from_stats_objects(
            ...     stats1, stats2)
            >>> assert stats_combined.unit == 1E-6
            >>> assert stats_combined.timings == {
            ...     ('foo', 1, 'spam.py'): [(2, 10, 300)],
            ...     ('bar', 10, 'spam.py'):
            ...     [(11, 12, 3000), (12, 6, 600)],
            ...     ('baz', 5, 'eggs.py'): [(5, 2, 500)]}
        """
        timings, unit = cls._get_aggregated_timings([stats, *more_stats])
        return cls(timings, unit)

    @staticmethod
    def _get_aggregated_timings(stats_objs):
        if not stats_objs:
            raise ValueError(f'stats_objs = {stats_objs!r}: empty')
        try:
            stats, = stats_objs
        except ValueError:  # > 1 obj
            # Add from small scaling factors to large to minimize
            # rounding errors
            stats_objs = sorted(stats_objs, key=operator.attrgetter('unit'))
            unit = stats_objs[-1].unit
            # type: dict[tuple[str, int, int], dict[int, tuple[int, float]]
            timing_dict = {}
            for stats in stats_objs:
                factor = stats.unit / unit
                for key, entries in stats.timings.items():
                    entry_dict = timing_dict.setdefault(key, {})
                    for lineno, nhits, time in entries:
                        prev_nhits, prev_time = entry_dict.get(lineno, (0, 0))
                        entry_dict[lineno] = (prev_nhits + nhits,
                                              prev_time + factor * time)
            timings = {
                key: [(lineno, nhits, int(round(time, 0)))
                      for lineno, (nhits, time) in sorted(entry_dict.items())]
                for key, entry_dict in timing_dict.items()}
        else:
            timings = {key: entries.copy()
                       for key, entries in stats.timings.items()}
            unit = stats.unit
        return timings, unit


class LineProfiler(CLineProfiler, ByCountProfilerMixin):
    """
    A profiler that records the execution times of individual lines.

    This provides the core line-profiler functionality.

    Example:
        >>> import line_profiler
        >>> profile = line_profiler.LineProfiler()
        >>> @profile
        ... def func():
        ...     x1 = list(range(10))
        ...     x2 = list(range(100))
        ...     x3 = list(range(1000))
        >>> func()
        >>> profile.print_stats()
    """
    def __call__(self, func):
        """
        Decorate a function, method, :py:class:`property`,
        :py:func:`~functools.partial` object etc. to start the profiler
        on function entry and stop it on function exit.
        """
        # The same object is returned when:
        # - `func` is a `types.FunctionType` which is already
        #   decorated by the profiler,
        # - `func` is a class, or
        # - `func` is any of the C-level callables that can't be
        #   profiled
        # otherwise, wrapper objects are always returned.
        self.add_callable(func)
        return self.wrap_callable(func)

    def wrap_callable(self, func):
        if is_c_level_callable(func):  # Non-profilable
            return func
        return super().wrap_callable(func)

    def add_callable(self, func, guard=None, name=None):
        """
        Register a function, method, :py:class:`property`,
        :py:func:`~functools.partial` object, etc. with the underlying
        Cython profiler.

        Args:
            func (...):
                Function, class/static/bound method, property, etc.
            guard (Optional[Callable[[types.FunctionType], bool]])
                Optional checker callable, which takes a function object
                and returns true(-y) if it *should not* be passed to
                :py:meth:`.add_function()`.  Defaults to checking
                whether the function is already a profiling wrapper.
            name (Optional[str])
                Optional name for ``func``, to be used in log messages.

        Returns:
            1 if any function is added to the profiler, 0 otherwise.

        Note:
            This method should in general be called instead of the more
            low-level :py:meth:`.add_function()`.
        """
        if guard is None:
            guard = self._already_a_wrapper

        nadded = 0
        func_repr = self._repr_for_log(func, name)
        for impl in self.get_underlying_functions(func):
            info, wrapped_by_this_prof = self._get_wrapper_info(impl)
            if wrapped_by_this_prof if guard is None else guard(impl):
                continue
            if info:
                # It's still a profiling wrapper, just wrapped by
                # someone else -> extract the inner function
                impl = info.func
            self.add_function(impl)
            nadded += 1
            if impl is func:
                self._debug(f'added {func_repr}')
            else:
                self._debug(f'added {func_repr} -> {self._repr_for_log(impl)}')

        return 1 if nadded else 0

    @staticmethod
    def _repr_for_log(obj, name=None):
        try:
            real_name = '{0.__module__}.{0.__qualname__}'.format(obj)
        except AttributeError:
            try:
                real_name = obj.__name__
            except AttributeError:
                real_name = '???'
        return '{} `{}{}` {}@ {:#x}'.format(
            type(obj).__name__,
            real_name,
            '()' if callable(obj) and not isinstance(obj, type) else '',
            f'(=`{name}`) ' if name and name != real_name else '',
            id(obj))

    def _debug(self, msg):
        self_repr = f'{type(self).__name__} @ {id(self):#x}'
        logger = diagnostics.log
        if logger.backend == 'print':
            now = datetime.now().isoformat(sep=' ', timespec='seconds')
            msg = f'[{self_repr} {now}] {msg}'
        else:
            msg = f'{self_repr}: {msg}'
        logger.debug(msg)

    def get_stats(self):
        return LineStats.from_stats_objects(super().get_stats())

    def dump_stats(self, filename):
        """ Dump a representation of the data to a file as a pickled
        :py:class:`~.LineStats` object from :py:meth:`~.get_stats()`.
        """
        self.get_stats().to_file(filename)

    def print_stats(self, stream=None, output_unit=None, stripzeros=False,
                    details=True, summarize=False, sort=False, rich=False, *,
                    config=None):
        """ Show the gathered statistics.
        """
        self.get_stats().print(
            stream=stream, output_unit=output_unit,
            stripzeros=stripzeros, details=details, summarize=summarize,
            sort=sort, rich=rich, config=config)

    def _add_namespace(
            self, namespace, *,
            seen=None,
            func_scoping_policy=ScopingPolicy.NONE,
            class_scoping_policy=ScopingPolicy.NONE,
            module_scoping_policy=ScopingPolicy.NONE,
            wrap=False,
            name=None):
        def func_guard(func):
            return self._already_a_wrapper(func) or not func_check(func)

        if seen is None:
            seen = set()
        count = 0
        add_namespace = functools.partial(
            self._add_namespace,
            seen=seen,
            func_scoping_policy=func_scoping_policy,
            class_scoping_policy=class_scoping_policy,
            module_scoping_policy=module_scoping_policy,
            wrap=wrap)
        members_to_wrap = {}
        func_check = func_scoping_policy.get_filter(namespace, 'func')
        cls_check = class_scoping_policy.get_filter(namespace, 'class')
        mod_check = module_scoping_policy.get_filter(namespace, 'module')

        # Logging stuff
        if not name:
            try:  # Class
                name = '{0.__module__}.{0.__qualname__}'.format(namespace)
            except AttributeError:  # Module
                name = namespace.__name__

        for attr, value in vars(namespace).items():
            if id(value) in seen:
                continue
            seen.add(id(value))
            if isinstance(value, type):
                if not (cls_check(value)
                        and add_namespace(value, name=f'{name}.{attr}')):
                    continue
            elif isinstance(value, types.ModuleType):
                if not (mod_check(value)
                        and add_namespace(value, name=f'{name}.{attr}')):
                    continue
            else:
                try:
                    if not self.add_callable(
                            value, guard=func_guard, name=f'{name}.{attr}'):
                        continue
                except TypeError:  # Not a callable (wrapper)
                    continue
                if wrap:
                    members_to_wrap[attr] = value
            count += 1
        if wrap and members_to_wrap:
            self._wrap_namespace_members(namespace, members_to_wrap,
                                         warning_stack_level=3)
        if count:
            self._debug(
                'added {} member{} in {}'.format(
                    count,
                    '' if count == 1 else 's',
                    self._repr_for_log(namespace, name)))
        return count

    def add_class(self, cls, *, scoping_policy=None, wrap=False):
        """
        Add the members (callables (wrappers), methods, classes, ...) in
        a class' local namespace and profile them.

        Args:
            cls (type):
                Class to be profiled.
            scoping_policy (Union[str, ScopingPolicy, \
ScopingPolicyDict, None]):
                Whether (and how) to match the scope of members and
                decide on whether to add them:

                :py:class:`str` (incl. :py:class:`~.ScopingPolicy`):
                    Strings are converted to :py:class:`~.ScopingPolicy`
                    instances in a case-insensitive manner, and the same
                    policy applies to all members.

                ``{'func': ..., 'class': ..., 'module': ...}``
                    Mapping specifying individual policies to be enacted
                    for the corresponding member types.

                :py:const:`None`
                    The default, equivalent to
                    :py:data:\
`~.scoping_policy.DEFAULT_SCOPING_POLICIES`.

                See :py:class:`~.ScopingPolicy` and
                :py:meth:`.ScopingPolicy.to_policies` for details.
            wrap (bool):
                Whether to replace the wrapped members with wrappers
                which automatically enable/disable the profiler when
                called.

        Returns:
            n (int):
                Number of members added to the profiler.
        """
        policies = ScopingPolicy.to_policies(scoping_policy)
        return self._add_namespace(cls,
                                   func_scoping_policy=policies['func'],
                                   class_scoping_policy=policies['class'],
                                   module_scoping_policy=policies['module'],
                                   wrap=wrap)

    def add_module(self, mod, *, scoping_policy=None, wrap=False):
        """
        Add the members (callables (wrappers), methods, classes, ...) in
        a module's local namespace and profile them.

        Args:
            mod (ModuleType):
                Module to be profiled.
            scoping_policy (Union[str, ScopingPolicy, \
ScopingPolicyDict, None]):
                Whether (and how) to match the scope of members and
                decide on whether to add them:

                :py:class:`str` (incl. :py:class:`~.ScopingPolicy`):
                    Strings are converted to :py:class:`~.ScopingPolicy`
                    instances in a case-insensitive manner, and the same
                    policy applies to all members.

                ``{'func': ..., 'class': ..., 'module': ...}``
                    Mapping specifying individual policies to be enacted
                    for the corresponding member types.

                :py:const:`None`
                    The default, equivalent to
                    :py:data:\
`~.scoping_policy.DEFAULT_SCOPING_POLICIES`.

                See :py:class:`~.ScopingPolicy` and
                :py:meth:`.ScopingPolicy.to_policies` for details.
            wrap (bool):
                Whether to replace the wrapped members with wrappers
                which automatically enable/disable the profiler when
                called.

        Returns:
            n (int):
                Number of members added to the profiler.
        """
        policies = ScopingPolicy.to_policies(scoping_policy)
        return self._add_namespace(mod,
                                   func_scoping_policy=policies['func'],
                                   class_scoping_policy=policies['class'],
                                   module_scoping_policy=policies['module'],
                                   wrap=wrap)

    def _get_wrapper_info(self, func):
        info = getattr(func, self._profiler_wrapped_marker, None)
        return info, bool(info and id(self) == info.profiler_id)

    # Override these mixed-in bookkeeping methods to take care of
    # potential multiple profiler sequences

    def _already_a_wrapper(self, func):
        return self._get_wrapper_info(func)[1]

    def _mark_wrapper(self, wrapper):
        # Are re-wrapping an existing wrapper (e.g. created by another
        # profiler?)
        wrapped = wrapper.__wrapped__
        info = getattr(wrapped, self._profiler_wrapped_marker, None)
        new_info = _WrapperInfo(info.func if info else wrapped, id(self))
        setattr(wrapper, self._profiler_wrapped_marker, new_info)
        return wrapper


# This could be in the ipython_extension submodule,
# but it doesn't depend on the IPython module so it's easier to just let it stay here.
def is_generated_code(filename):
    """ Return True if a filename corresponds to generated code, such as a
    Jupyter Notebook cell.
    """
    filename = os.path.normcase(filename)
    temp_dir = os.path.normcase(tempfile.gettempdir())
    return (
        filename.startswith('<generated') or
        filename.startswith('<ipython-input-') or
        filename.startswith(os.path.join(temp_dir, 'ipykernel_')) or
        filename.startswith(os.path.join(temp_dir, 'xpython_'))
    )


def show_func(filename, start_lineno, func_name, timings, unit,
              output_unit=None, stream=None, stripzeros=False, rich=False,
              *,
              config=None):
    """
    Show results for a single function.

    Args:
        filename (str):
            Path to the profiled file

        start_lineno (int):
            First line number of profiled function

        func_name (str): name of profiled function

        timings (List[Tuple[int, int, float]]):
            Measurements for each line (lineno, nhits, time).

        unit (float):
            The number of seconds used as the cython LineProfiler's unit.

        output_unit (float | None):
            Output unit (in seconds) in which the timing info is displayed.

        stream (io.TextIOBase | None):
            Defaults to sys.stdout

        stripzeros (bool):
            If True, prints nothing if the function was not run

        rich (bool):
            If True, attempt to use rich highlighting.

        config (Union[str, PurePath, bool, None]):
            Optional filename from which to load configurations (e.g.
            output column widths);
            default (= `True` or `None`) is to look for a config file
            based on the environment variable `${LINE_PROFILER_RC}` and
            path-based lookup;
            passing `False` disables all lookup and falls back to the
            default configuration

    Example:
        >>> from line_profiler.line_profiler import show_func
        >>> import line_profiler
        >>> # Use a function in this file as an example
        >>> func = line_profiler.line_profiler.show_text
        >>> start_lineno = func.__code__.co_firstlineno
        >>> filename = func.__code__.co_filename
        >>> func_name = func.__name__
        >>> # Build fake timeings for each line in the example function
        >>> import inspect
        >>> num_lines = len(inspect.getsourcelines(func)[0])
        >>> line_numbers = list(range(start_lineno + 3,
        ...                           start_lineno + num_lines))
        >>> timings = [(lineno, idx * 1e13, idx * (2e10 ** (idx % 3)))
        ...            for idx, lineno
        ...            in enumerate(line_numbers, start=1)]
        >>> unit = 1.0
        >>> output_unit = 1.0
        >>> stream = None
        >>> stripzeros = False
        >>> rich = 1
        >>> show_func(filename, start_lineno, func_name, timings, unit,
        ...           output_unit, stream, stripzeros, rich)
    """
    if stream is None:
        stream = sys.stdout

    total_hits = sum(t[1] for t in timings)
    total_time = sum(t[2] for t in timings)

    if stripzeros and total_hits == 0:
        return

    if rich:
        # References:
        # https://github.com/Textualize/rich/discussions/3076
        try:
            from rich.syntax import Syntax
            from rich.highlighter import ReprHighlighter
            from rich.text import Text
            from rich.console import Console
            from rich.table import Table
        except ImportError:
            rich = 0

    if output_unit is None:
        output_unit = unit
    scalar = unit / output_unit

    linenos = [t[0] for t in timings]

    stream.write('Total time: %g s\n' % (total_time * unit))
    if os.path.exists(filename) or is_generated_code(filename):
        stream.write(f'File: {filename}\n')
        stream.write(f'Function: {func_name} at line {start_lineno}\n')
        if os.path.exists(filename):
            # Clear the cache to ensure that we get up-to-date results.
            linecache.clearcache()
        sublines = get_code_block(filename, start_lineno)
    else:
        stream.write('\n')
        stream.write(f'Could not find file {filename}\n')
        stream.write('Are you sure you are running this program from the same directory\n')
        stream.write('that you ran the profiler from?\n')
        stream.write("Continuing without the function's contents.\n")
        # Fake empty lines so we can see the timings, if not the code.
        nlines = 1 if not linenos else max(linenos) - min(min(linenos), start_lineno) + 1
        sublines = [''] * nlines

    # Define minimum column sizes so text fits and usually looks consistent
    conf_column_sizes = get_column_widths(config)
    default_column_sizes = {
        col: max(width, conf_column_sizes.get(col, width))
        for col, width in get_column_widths().items()}

    display = {}

    # Loop over each line to determine better column formatting.
    # Fallback to scientific notation if columns are larger than a threshold.
    for lineno, nhits, time in timings:
        if total_time == 0:  # Happens rarely on empty function
            percent = ''
        else:
            percent = '%5.1f' % (100 * time / total_time)

        time_disp = '%5.1f' % (time * scalar)
        if len(time_disp) > default_column_sizes['time']:
            time_disp = '%5.3g' % (time * scalar)

        perhit_disp = '%5.1f' % (float(time) * scalar / nhits)
        if len(perhit_disp) > default_column_sizes['perhit']:
            perhit_disp = '%5.3g' % (float(time) * scalar / nhits)

        nhits_disp = "%d" % nhits
        if len(nhits_disp) > default_column_sizes['hits']:
            nhits_disp = '%g' % nhits

        display[lineno] = (nhits_disp, time_disp, perhit_disp, percent)

    # Expand column sizes if the numbers are large.
    column_sizes = default_column_sizes.copy()
    if len(display):
        max_hitlen = max(len(t[0]) for t in display.values())
        max_timelen = max(len(t[1]) for t in display.values())
        max_perhitlen = max(len(t[2]) for t in display.values())
        column_sizes['hits'] = max(column_sizes['hits'], max_hitlen)
        column_sizes['time'] = max(column_sizes['time'], max_timelen)
        column_sizes['perhit'] = max(column_sizes['perhit'], max_perhitlen)

    col_order = ['line', 'hits', 'time', 'perhit', 'percent']
    lhs_template = ' '.join(['%' + str(column_sizes[k]) + 's' for k in col_order])
    template = lhs_template + '  %-s'

    linenos = range(start_lineno, start_lineno + len(sublines))
    empty = ('', '', '', '')
    header = ('Line #', 'Hits', 'Time', 'Per Hit', '% Time', 'Line Contents')
    header = template % header
    stream.write('\n')
    stream.write(header)
    stream.write('\n')
    stream.write('=' * len(header))
    stream.write('\n')

    if rich:
        # Build the RHS and LHS of the table separately
        lhs_lines = []
        rhs_lines = []
        for lineno, line in zip(linenos, sublines):
            nhits, time, per_hit, percent = display.get(lineno, empty)
            txt = lhs_template % (lineno, nhits, time, per_hit, percent)
            rhs_lines.append(line.rstrip('\n').rstrip('\r'))
            lhs_lines.append(txt)

        rhs_text = '\n'.join(rhs_lines)
        lhs_text = '\n'.join(lhs_lines)

        # Highlight the RHS with Python syntax
        rhs = Syntax(rhs_text, 'python', background_color='default')

        # Use default highlights for the LHS
        # TODO: could use colors to draw the eye to longer running lines.
        lhs = Text(lhs_text)
        ReprHighlighter().highlight(lhs)

        # Use a table to horizontally concatenate the text
        # reference: https://github.com/Textualize/rich/discussions/3076
        table = Table(box=None,
                      padding=0,
                      collapse_padding=True,
                      show_header=False,
                      show_footer=False,
                      show_edge=False,
                      pad_edge=False,
                      expand=False)
        table.add_row(lhs, '  ', rhs)

        # Use a Console to render to the stream
        # Not sure if we should force-terminal or just specify the color system
        # write_console = Console(file=stream, force_terminal=True, soft_wrap=True)
        write_console = Console(file=stream, soft_wrap=True, color_system='standard')
        write_console.print(table)
        stream.write('\n')
    else:
        for lineno, line in zip(linenos, sublines):
            nhits, time, per_hit, percent = display.get(lineno, empty)
            line_ = line.rstrip('\n').rstrip('\r')
            txt = template % (lineno, nhits, time, per_hit, percent, line_)
            try:
                stream.write(txt)
            except UnicodeEncodeError:
                # todo: better handling of windows encoding issue
                # for now just work around it
                line_ = 'UnicodeEncodeError - help wanted for a fix'
                txt = template % (lineno, nhits, time, per_hit, percent, line_)
                stream.write(txt)

            stream.write('\n')
    stream.write('\n')


def show_text(stats, unit, output_unit=None, stream=None, stripzeros=False,
              details=True, summarize=False, sort=False, rich=False, *,
              config=None):
    """
    Show text for the given timings.

    Ignore:
        # For developer testing, generate some profile output
        python -m kernprof -l -p uuid -m uuid

        # Use this function to view it with rich
        python -m line_profiler -rmtz "uuid.lprof"

        # Use this function to view it without rich
        python -m line_profiler -mtz "uuid.lprof"
    """
    if stream is None:
        stream = sys.stdout

    if output_unit is not None:
        stream.write('Timer unit: %g s\n\n' % output_unit)
    else:
        stream.write('Timer unit: %g s\n\n' % unit)

    if sort:
        # Order by ascending duration
        stats_order = sorted(stats.items(), key=lambda kv: sum(t[2] for t in kv[1]))
    else:
        # Default ordering
        stats_order = stats.items()

    # Pre-lookup the appropriate config file
    config = ConfigSource.from_config(config).path

    if details:
        # Show detailed per-line information for each function.
        for (fn, lineno, name), timings in stats_order:
            show_func(fn, lineno, name, stats[fn, lineno, name], unit,
                      output_unit=output_unit, stream=stream,
                      stripzeros=stripzeros, rich=rich, config=config)

    if summarize:
        # Summarize the total time for each function
        if rich:
            try:
                from rich.console import Console
                from rich.markup import escape
            except ImportError:
                rich = 0
        line_template = '%6.2f seconds - %s:%s - %s'
        if rich:
            write_console = Console(file=stream, soft_wrap=True,
                                    color_system='standard')
            for (fn, lineno, name), timings in stats_order:
                total_time = sum(t[2] for t in timings) * unit
                if not stripzeros or total_time:
                    # Wrap the filename with link markup to allow the user to
                    # open the file
                    fn_link = f'[link={fn}]{escape(fn)}[/link]'
                    line = line_template % (total_time, fn_link, lineno, escape(name))
                    write_console.print(line)
        else:
            for (fn, lineno, name), timings in stats_order:
                total_time = sum(t[2] for t in timings) * unit
                if not stripzeros or total_time:
                    line = line_template % (total_time, fn, lineno, name)
                    stream.write(line + '\n')


load_stats = LineStats.from_files


def main():
    """
    The line profiler CLI to view output from :command:`kernprof -l`.
    """
    parser = ArgumentParser(
        description='Read and show line profiling results (`.lprof` files) '
        'as generated by the CLI application `kernprof` or by '
        '`LineProfiler.dump_stats()`.')
    get_main_config = functools.partial(get_cli_config, 'cli')
    default = config = get_main_config()

    add_argument(parser, '-V', '--version',
                 action='version', version=__version__)
    add_argument(parser, '-c', '--config',
                 help='Path to the TOML file, from the '
                 '`tool.line_profiler.cli` table of which to load '
                 'defaults for the options. '
                 f'(Default: {short_string_path(default.path)!r})')
    add_argument(parser, '--no-config',
                 action='store_const', dest='config', const=False,
                 help='Disable the loading of configuration files other than '
                 'the default one')
    add_argument(parser, '-u', '--unit', type=positive_float,
                 help='Output unit (in seconds) in which '
                 'the timing info is displayed. '
                 f'(Default: {default.conf_dict["unit"]} s)')
    add_argument(parser, '-r', '--rich', action='store_true',
                 help='Use rich formatting. '
                 f'(Default: {default.conf_dict["rich"]})')
    add_argument(parser, '-z', '--skip-zero', action='store_true',
                 help='Hide functions which have not been called. '
                 f'(Default: {default.conf_dict["skip_zero"]})')
    add_argument(parser, '-t', '--sort', action='store_true',
                 help='Sort by ascending total time. '
                 f'(Default: {default.conf_dict["sort"]})')
    add_argument(parser, '-m', '--summarize', action='store_true',
                 help='Print a summary of total function time. '
                 f'(Default: {default.conf_dict["summarize"]})')
    add_argument(parser, 'profile_output',
                 nargs='+',
                 help="'*.lprof' file(s) created by `kernprof`")

    args = parser.parse_args()
    if args.config:
        config = get_main_config(args.config)
        args.config = config.path
    for key, default in config.conf_dict.items():
        if getattr(args, key, None) is None:
            setattr(args, key, default)

    lstats = LineStats.from_files(*args.profile_output)
    show_text(lstats.timings, lstats.unit,
              output_unit=args.unit,
              stripzeros=args.skip_zero,
              rich=args.rich,
              sort=args.sort,
              summarize=args.summarize,
              config=args.config)


if __name__ == '__main__':
    main()
